# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex climate networks.
"""

from typing import Tuple
from collections.abc import Hashable

import numpy as np
from tqdm import trange

from .climate_data import ClimateData
from .climate_network import ClimateNetwork


class HavlinClimateNetwork(ClimateNetwork):

    """
    Encapsulates a Havlin climate network.

    The similarity matrix associated with a Havlin climate network is the
    maximum-lag correlation matrix with each entry normalized by the
    cross-correlation function's standard deviation.

    Havlin climate networks are undirected so far.

    Havlin climate networks were studied for daily data in [Yamasaki2008]_,
    [Gozolchiani2008]_, [Yamasaki2009]_.

    .. note::
       So far, the cross-correlation functions are estimated using \
       convolution in Fourier space (FFT). This may not be reliable \
       for larger delays.
    """

    #
    #  Defines internal methods
    #

    def __init__(self, data, max_delay, threshold=None, link_density=None,
                 non_local=False, node_weight_type="surface", silence_level=0):
        """
        Initialize an instance of HavlinClimateNetwork.

        .. note::
           Either threshold **OR** link_density have to be given!

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos**2 lat)

        :type data: :class:`.ClimateData`
        :arg data: The climate data used for network construction.
        :arg float threshold: The threshold of similarity measure, above which
            two nodes are linked in the network.
        :arg float link_density: The networks's desired link density.
        :param int max_delay: Maximum delay for cross-correlation functions.
        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        :arg str node_weight_type: The type of geographical node weight to be
            used.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if silence_level <= 1:
            print("Generating a Havlin climate network...")
        self.silence_level = silence_level

        #  Set instance variables
        self._max_delay = 0
        self._correlation_lag = None
        assert isinstance(data, ClimateData)
        self.data: ClimateData = data
        """The climate data used for network construction."""
        self.N = data.grid.N
        self._prescribed_link_density = link_density

        self._mut_clim: int = 0
        self._set_max_delay(max_delay)
        ClimateNetwork.__init__(self, grid=self.data.grid,
                                similarity_measure=self._similarity_measure,
                                threshold=threshold,
                                link_density=link_density,
                                non_local=non_local,
                                directed=False,
                                node_weight_type=node_weight_type,
                                silence_level=silence_level)

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        return ClimateNetwork.__cache_state__(self)

    def __rec_cache_state__(self) -> Tuple[object, ...]:
        return ClimateNetwork.__rec_cache_state__(self) + (self.data,)

    def __str__(self):
        """
        Return a string version of the instance of HavlinClimateNetwork.
        """
        return f'HavlinClimateNetwork:\n \
                {ClimateNetwork.__str__(self)}\n \
                Maximum delay used for correlation strength estimation: \
                {self.get_max_delay()}'

    def clear_cache(self):
        """
        Clean up cache.
        """
        try:
            del self._correlation_lag
        except AttributeError:
            pass

    #
    #  Defines methods to calculate correlation strength and lags
    #

    #  TODO: Implement an algorithm, which is NOT based on FFT!
    def _calculate_correlation_strength(self, anomaly, max_delay, gamma=0.2):
        """
        Calculate correlation strength and maximum lag matrices.

        Follows the method described in [Yamasaki2008]_.

        Also returns the time lag at maximum correlation for each link.

        :type anomaly: 2D array [time, index]
        :arg anomaly: The anomaly data for network construction.
        :arg int max_delay: The maximum delay for cross-correlation functions.
        :arg float gamma: The width of decay region in cosine shaped window
            used for FFT cross-correlation estimation.
        :rtype: tuple of two 2D arrays [index, index]
        :return: the correlation strength and maximum lag matrices.
        """
        N = self.N
        self.data.normalize_time_series_array(anomaly)
        anomaly *= self.data.cos_window(anomaly, gamma)
        #  Zero pad windowed data to set the length of each time series to
        #  a power of two
        anomaly = self.data.zero_pad_data(anomaly)
        correlation_strength = np.empty((N, N))
        max_lag_matrix = np.empty((N, N))

        #  Calculate the inverse Fourier transform of all time series
        ifft = np.fft.ifft(anomaly, axis=0)

        for i in trange(N, disable=self.silence_level > 1):
            #  Calculate the cross correlation function of node i to all other
            #  nodes which is not normalized yet.
            #  The real part has to be taken to get rid of small imaginary
            #  parts due to rounding errors.
            cc_one_to_all = np.conjugate(np.tile(ifft[:, i],
                                                 (N, 1)).transpose()) * ifft
            cc_one_to_all = np.real(np.fft.fft(cc_one_to_all, axis=0))

            #  Consider only the cross correlations between -max_delay and
            #  +max_delay.
            #  Correlation values at negative lag are now stored left of the
            #  correlation values at positive lag.
            cc_one_to_all = np.concatenate((cc_one_to_all[-max_delay:-1, :],
                                            cc_one_to_all[0:max_delay, :]))

            #  Consider only absolute values
            cc_one_to_all = np.abs(cc_one_to_all)

            #  Store correlation strengths
            correlation_strength[i, :] = \
                cc_one_to_all.max(axis=0) / cc_one_to_all.std(axis=0)

            #  Store time delays at maximum cross correlation
            max_lag_matrix[i, :] = cc_one_to_all.argmax(axis=0) - max_delay

        return (correlation_strength, max_lag_matrix)

    def get_max_delay(self):
        """
        Return the maximum delay used for cross-correlation estimation.

        :return float: the maximum delay used for cross-correlation estimation.
        """
        return self._max_delay

    def _set_max_delay(self, max_delay):
        """
        Set the maximum lag time used for cross-correlation estimation.

        :arg int max_delay: The maximum delay for cross-correlation functions.
        """
        self._mut_clim += 1
        self._max_delay = max_delay
        results = self._calculate_correlation_strength(self.data.anomaly(),
                                                       max_delay)
        self._similarity_measure = results[0]
        self._correlation_lag = results[1]

    def set_max_delay(self, max_delay):
        """
        Set the maximum lag time used for cross-correlation estimation.

        (Re)generates the current Havlin climate network accordingly.

        :arg int max_delay: The maximum delay for cross-correlation functions.
        """
        self._set_max_delay(max_delay)
        self._regenerate_network()

    def correlation_strength(self):
        """
        Return the correlation strength matrix.

        :rtype: 2D array [index, index]
        :return: the correlation strength matrix.
        """
        return self._similarity_measure

    def correlation_lag(self):
        """
        Return the lag at maximum cross-correlation matrix.

        :rtype: 2D array [index, index]
        :return: the lag at maximum cross-correlation matrix.
        """
        return self._correlation_lag

    #
    #  Methods to calculate weighted network measures
    #

    def correlation_strength_weighted_average_path_length(self):
        """
        Return correlation strength weighted average path length.

        :return float: the correlation strength weighted average path length.
        """
        return self._weighted_metric(
            "correlation_strength",
            lambda: np.abs(self.correlation_strength()),
            "average_path_length")

    def correlation_strength_weighted_closeness(self):
        """
        Return correlation strength weighted closeness.

        :rtype: 1D array [index]
        :return: the correlation strength weighted closeness sequence.
        """
        return self._weighted_metric(
            "correlation_strength",
            lambda: np.abs(self.correlation_strength()),
            "closeness")

    def correlation_lag_weighted_average_path_length(self):
        """
        Return correlation lag weighted average path length.

        :return float: the correlation lag weighted average path length.
        """
        return self._weighted_metric(
            "correlation_lag",
            lambda: np.abs(self.correlation_lag()),
            "average_path_length")

    def correlation_lag_weighted_closeness(self):
        """
        Return correlation lag weighted closeness.

        :rtype: 1D array [index]
        :return: the correlation lag weighted closeness sequence.
        """
        return self._weighted_metric(
            "correlation_lag",
            lambda: np.abs(self.correlation_lag()),
            "closeness")

    def local_correlation_strength_weighted_vulnerability(self):
        """
        Return correlation strength weighted vulnerability.

        :rtype: 1D array [index]
        :return: the correlation strength weighted vulnerability sequence.
        """
        return self._weighted_metric(
            "correlation_strength",
            lambda: np.abs(self.correlation_strength()),
            "local_vulnerability")

    def local_correlation_lag_weighted_vulnerability(self):
        """
        Return correlation lag weighted vulnerability.

        :rtype: 1D array [index]
        :return: the correlation lag weighted vulnerability sequence.
        """
        return self._weighted_metric(
            "correlation_lag",
            lambda: np.abs(self.correlation_lag()),
            "local_vulnerability")



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Simple tests for the timeseries class.
"""

import pytest
import numpy as np
from numpy.testing import assert_array_almost_equal

from pyunicorn.timeseries import CrossRecurrencePlot, \
    InterSystemRecurrenceNetwork, Surrogates, VisibilityGraph
from pyunicorn.core.data import Data
from pyunicorn.core._ext.types import DFIELD


def create_test_data():
    # Create test time series
    tdata = Data.SmallTestData().observable()
    # subtract means form the input data
    tdata -= np.mean(tdata, axis=1)[:, None]
    # normalize the data
    tdata /= np.sqrt(np.sum(tdata*tdata, axis=1))[:, None]
    return tdata


# -----------------------------------------------------------------------------
# cross_recurrence_plot
# -----------------------------------------------------------------------------


@pytest.mark.parametrize("thresh, rr", [(.2, None), (None, .2)])
def testCrossRecurrencePlot(thresh, rr, metric: str):
    # create two instances of the same test dataset
    tdata1 = create_test_data()
    x1 = tdata1[:, 0]
    y1 = tdata1[:, 1]
    tdata2 = create_test_data()
    x2 = tdata2[:, 0]
    y2 = tdata2[:, 1]
    # create CrossRecurrencePlot for both
    crp1 = CrossRecurrencePlot(
            x1, y1, threshold=thresh, recurrence_rate=rr, metric=metric)
    crp2 = CrossRecurrencePlot(
            x2, y2, threshold=thresh, recurrence_rate=rr, metric=metric)
    # get respective distance matrices
    dist_1 = crp1.distance_matrix(metric)
    dist_2 = crp2.distance_matrix(metric)
    # get respective recurrence matrices
    CR1 = crp1.recurrence_matrix()
    CR2 = crp2.recurrence_matrix()

    assert np.allclose(dist_1, dist_2, atol=1e-04)
    assert CR1.shape == CR2.shape
    assert CR1.shape == (len(x1), len(y1))
    assert CR1.dtype == CR2.dtype
    assert CR1.dtype == np.int8


# -----------------------------------------------------------------------------
# inter_system_recurrence_network
# -----------------------------------------------------------------------------


@pytest.mark.parametrize("thresh, rr",
                         [((.2, .3, .2), None), (None, (.2, .3, .2))], ids=str)
def testInterSystemRecurrenceNetwork(thresh, rr, metric: str):
    # create two instances of the same test dataset
    tdata1 = create_test_data()
    x1 = tdata1[:, 0]
    y1 = tdata1[:, 1]
    tdata2 = create_test_data()
    x2 = tdata2[:, 0]
    y2 = tdata2[:, 1]
    # create InterSystemRecurrenceNetwork for both
    isrn1 = InterSystemRecurrenceNetwork(
            x1, y1, threshold=thresh, recurrence_rate=rr, metric=metric)
    isrn2 = InterSystemRecurrenceNetwork(
            x2, y2, threshold=thresh, recurrence_rate=rr, metric=metric)
    # get respective adjacency matrices
    A1 = isrn1.adjacency
    A2 = isrn2.adjacency

    assert np.array_equal(A1, A2)
    assert A1.shape == A2.shape
    assert A1.shape == (len(x1)*2, len(y1)*2)
    assert A1.dtype == A2.dtype
    assert A1.dtype == np.int16


# -----------------------------------------------------------------------------
# surrogates
# -----------------------------------------------------------------------------


def testNormalizeTimeSeriesArray():
    ts = Surrogates.SmallTestData().original_data
    Surrogates.SmallTestData().normalize_time_series_array(ts)
    res = ts.mean(axis=1)
    exp = np.array([0., 0., 0., 0., 0., 0.])
    assert np.allclose(res, exp, atol=1e-04)

    res = ts.std(axis=1)
    exp = np.array([1., 1., 1., 1., 1., 1.])
    assert np.allclose(res, exp, atol=1e-04)


def testEmbedTimeSeriesArray():
    ts = Surrogates.SmallTestData().original_data
    res = Surrogates.SmallTestData().embed_time_series_array(
        time_series_array=ts, dimension=3, delay=2)[0, :6, :]
    exp = np.array([[0., 0.61464833, 1.14988147],
                    [0.31244015, 0.89680225, 1.3660254],
                    [0.61464833, 1.14988147, 1.53884177],
                    [0.89680225, 1.3660254, 1.6636525],
                    [1.14988147, 1.53884177, 1.73766672],
                    [1.3660254, 1.6636525, 1.76007351]])
    assert np.allclose(res, exp, atol=1e-04)


def testWhiteNoiseSurrogates():
    ts = Surrogates.SmallTestData().original_data
    surrogates = Surrogates.SmallTestData().white_noise_surrogates(ts)

    assert np.allclose(np.histogram(ts[0, :])[0],
                       np.histogram(surrogates[0, :])[0])


def testCorrelatedNoiseSurrogates():
    ts = Surrogates.SmallTestData().original_data
    surrogates = Surrogates.SmallTestData().correlated_noise_surrogates(ts)
    assert np.allclose(np.abs(np.fft.fft(ts, axis=1))[0, 1:10],
                       np.abs(np.fft.fft(surrogates, axis=1))[0, 1:10])


def testTwinSurrogates():
    tdata = create_test_data()
    n_index, n_times = tdata.shape
    s = Surrogates(tdata)
    tsurro = s.twin_surrogates(tdata, 1, 0, 0.2)
    corrcoef = np.corrcoef(tdata, tsurro)[n_index:, :n_index]
    for i in range(n_index):
        corrcoef[i, i] = 0.0
    assert (corrcoef >= -1.0).all() and (corrcoef <= 1.0).all()


def testPearsonCorrelation():
    tdata = create_test_data()
    n_index, n_times = tdata.shape
    norm = 1.0 / float(n_times)
    c = Surrogates.test_pearson_correlation(tdata, tdata)
    corrcoef = np.corrcoef(tdata, tdata)[n_index:, :n_index]*norm
    for i in range(n_index):
        corrcoef[i, i] = 0.0

    assert c.shape == (n_index, n_index)
    assert_array_almost_equal(c, corrcoef, decimal=5)


def testMutualInformation():
    tdata = create_test_data()
    test_mi = Surrogates.test_mutual_information(tdata[:1], tdata[-1:],
                                                 n_bins=32)
    assert (test_mi >= -1.0).all() and (test_mi <= 1.0).all()

# -----------------------------------------------------------------------------
# visibility_graph
# -----------------------------------------------------------------------------


def create_test_timeseries():
    """
    Return test time series including random values and timings
    with 10 sampling points.

    :rtype: 2D array
    :return: a timeseries for testing purposes
    """
    #  Create time series
    ts = np.zeros((2, 10))

    for i in range(2):
        ts[i, :] = np.random.rand(10)

    ts[1, :].sort()

    return ts


# def testVisibility():
#     x, t = create_test_timeseries()
#     n_times = len(t)
#     vg = VisibilityGraph(x, timings=t)
#     # Choose two different, not neighbouring random nodes i, j
#     node1, node2 = 0, 0
#     while (abs(node2-node1)<=1):
#         node1 = np.int32(np.floor(np.random.rand()*n_times))
#         node2 = np.int32(np.floor(np.random.rand()*n_times))

#     # the following returns incorrect results, assertion therefore failing
#     i, j = min(node1, node2), max(node1, node2)
#     test = np.zeros((j-(i+1)))
#     for k in range(i+1,j):
#         test[k-(i+1)] = np.less((x[k]-x[i]) / (t[k]-t[i]),
#                                 (x[j]-x[i]) / (t[j]-t[i]))

#     assert np.invert(bool(np.sum(test))) == vg.visibility(node1, node2)


def testVisibility():
    x, t = create_test_timeseries()
    vg = VisibilityGraph(x, timings=t)
    A = vg.adjacency

    assert A.shape == (len(x), len(x))
    assert A.dtype == np.int16


def testVisibilityHorizontal():
    x, t = create_test_timeseries()
    vg = VisibilityGraph(x, timings=t, horizontal=True)
    A = vg.adjacency

    assert A.shape == (len(x), len(x))
    assert A.dtype == np.int16


def testRetardedLocalClustering():
    x, t = create_test_timeseries()
    vg = VisibilityGraph(x, timings=t)
    C_ret = vg.retarded_local_clustering()

    assert C_ret.shape == x.shape
    assert C_ret.dtype == DFIELD


def testAdvancedLocalClustering():
    x, t = create_test_timeseries()
    vg = VisibilityGraph(x, timings=t)
    C_adv = vg.advanced_local_clustering()

    assert C_adv.shape == x.shape
    assert C_adv.dtype == DFIELD



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for the analysis of dynamical systems and time series based
on recurrence plots, including measures of recurrence quantification
analysis (RQA) and recurrence network analysis.
"""

import numpy as np

from .recurrence_plot import RecurrencePlot


class JointRecurrencePlot(RecurrencePlot):

    """
    Class JointRecurrencePlot for generating and quantitatively analyzing
    :index:`joint recurrence plots <single: joint recurrence plot>`.

    The JointRecurrencePlot class supports the construction of joint recurrence
    plots from two multi-dimensional time series, optionally using embedding.
    Currently, manhattan, euclidean and supremum norms are provided for
    measuring distances in phase space.

    Methods for calculating commonly used measures of :index:`recurrence
    quantification analysis <pair: RQA; joint recurrence plot>` (RQA) are
    provided, e.g., determinism, maximum diagonal line length and laminarity.
    The definitions of these measures together with a review of the theory and
    applications of joint recurrence plots are given in [Marwan2007]_.

    **Examples:**

     - Create an instance of JointRecurrencePlot with a fixed recurrence
       threshold and without embedding::

           JointRecurrencePlot(x, y, threshold=(0.1,0.2))

     - Create an instance of JointRecurrencePlot with a fixed recurrence
       threshold in units of STD and without embedding::

           JointRecurrencePlot(x, y, threshold_std=(0.03,0.05))

     - Create an instance of JointRecurrencePlot at a fixed recurrence rate and
       using time delay embedding::

           JointRecurrencePlot(
               x, y, dim=(3,5), tau=(2,1),
               recurrence_rate=(0.05,0.04)).recurrence_rate()
    """

    #
    #  Internal methods
    #

    def __init__(self, x, y, metric=("supremum", "supremum"),
                 normalize=False, lag=0, silence_level=0, **kwds):
        """
        Initialize an instance of JointRecurrencePlot.

        .. note::
           For a joint recurrence plot, time series x and y need to have the
           same length!

        Either recurrence thresholds ``threshold``/``threshold_std`` or
        recurrence rates ``recurrence_rate`` have to be given as keyword
        arguments.

        Embedding is only supported for scalar time series. If embedding
        dimension ``dim`` and delay ``tau`` are **both** given as keyword
        arguments, embedding is applied. Multidimensional time series are
        processed as is by default.

        :type x: 2D array (time, dimension)
        :arg x: Time series x to be analyzed (scalar/multi-dimensional).
        :type y: 2D array (time, dimension)
        :arg y: Time series y to be analyzed (scalar/multi-dimensional).
        :type metric: tuple of string
        :arg metric: The metric for measuring distances in phase space
            ("manhattan", "euclidean", "supremum"). Give separately for each
            time series.
        :type normalize: tuple of bool
        :arg normalize: Decide whether to normalize the time series to zero
            mean and unit standard deviation. Give separately for each time
            series.
        :arg number lag: To create a delayed version of the JRP.
        :arg number silence_level: Inverse level of verbosity of the object.
        :type threshold: tuple of number
        :keyword threshold: The recurrence threshold keyword for generating the
            recurrence plot using a fixed threshold.  Give separately for each
            time series.
        :type threshold_std: tuple of number
        :keyword threshold_std: The recurrence threshold keyword for generating
            the recurrence plot using a fixed threshold in units of the time
            series' STD. Give separately for each time series.
        :type recurrence_rate: tuple of number
        :keyword recurrence_rate: The recurrence rate keyword for generating
            the recurrence plot using a fixed recurrence rate. Give separately
            for each time series.
        :type dim: tuple of number
        :keyword dim: Embedding dimension. Give separately for each time
            series.
        :type tau: tuple of number
        :keyword tau: Embedding delay. Give separately for each time series.
        """
        threshold = kwds.get("threshold")
        threshold_std = kwds.get("threshold_std")
        recurrence_rate = kwds.get("recurrence_rate")

        assert len(metric) == 2
        assert all(m in ("manhattan", "euclidean", "supremum") for m in metric)

        # initialise base class using dummy values (bad OO design)
        RecurrencePlot.__init__(
            self, np.empty((2, 0)), metric=metric[0], normalize=normalize,
            threshold=threshold[0] if threshold else 0,
            recurrence_rate=recurrence_rate, silence_level=silence_level)

        #  Store type of metric
        self.metric = metric

        self.JR = None
        """The joint recurrence matrix."""
        self.N = 0
        """The length of both embedded time series x and y."""

        #  Check for consistency: x and y need to have the same length
        if x.shape[0] == y.shape[0]:

            #  Store time series
            self.x = x.copy().astype("float32")
            """The time series x."""
            self.y = y.copy().astype("float32")
            """The time series y."""

            #  Reshape time series
            self.x.shape = (self.x.shape[0], -1)
            self.y.shape = (self.y.shape[0], -1)

            #  Normalize time series
            if normalize:
                self.normalize_time_series(self.x)
                self.normalize_time_series(self.y)

            #  Store lag
            self.lag = lag
            """Used to create a delayed JRP."""

            #  Get embedding dimension and delay from **kwds
            dim = kwds.get("dim")
            tau = kwds.get("tau")

            if dim is not None and tau is not None:
                #  Embed the time series
                self.x_embedded = \
                    self.embed_time_series(self.x, dim[0], tau[0])
                """The embedded time series x."""
                self.y_embedded = \
                    self.embed_time_series(self.y, dim[1], tau[1])
                """The embedded time series y."""
            else:
                self.x_embedded = self.x
                self.y_embedded = self.y

            #  Prune embedded time series to same length
            #  (number of state vectors)
            min_N = min(self.x_embedded.shape[0], self.y_embedded.shape[0])
            self.x_embedded = self.x_embedded[:min_N, :]
            self.y_embedded = self.y_embedded[:min_N, :]

            #  construct recurrence plot accordingly to
            #  threshold / recurrence rate
            if np.abs(lag) > x.shape[0]:
                #  Lag must be smaller than size of recurrence plot
                raise ValueError("Delay value (lag) must not exceed length of \
                                 time series!")
            if threshold is not None:
                #  Calculate the recurrence matrix R using the radius of
                #  neighborhood threshold
                JointRecurrencePlot.set_fixed_threshold(self, threshold)
            elif threshold_std is not None:
                #  Calculate the recurrence matrix R using the radius of
                #  neighborhood threshold in units of the time series' STD
                JointRecurrencePlot.\
                    set_fixed_threshold_std(self, threshold_std)
            elif recurrence_rate is not None:
                #  Calculate the recurrence matrix R using a fixed recurrence
                #  rate
                JointRecurrencePlot.\
                    set_fixed_recurrence_rate(self, recurrence_rate)
            else:
                raise NameError("Please give either threshold or \
                                recurrence_rate to construct the joint \
                                recurrence plot!")

            #  No treatment of missing values yet!
            self.missing_values = False

        else:
            raise ValueError("Both time series x and y need to have the same \
                             length!")

    def __str__(self):
        """
        Returns a string representation.
        """
        return ("JointRecurrencePlot: "
                f"time series shapes {self.x.shape}.\n"
                f"Embedding dimension {self.dim if self.dim else 0}\n"
                f"Threshold {self.threshold}, {self.metric} metric")

    #
    #  Service methods
    #

    def recurrence_matrix(self):
        """
        Return the current joint recurrence matrix :math:`JR`.

        :rtype: 2D square Numpy array
        :return: the current joint recurrence matrix :math:`JR`.
        """
        return self.JR

    #
    #  Calculate recurrence plot
    #

    def set_fixed_threshold(self, threshold):
        """
        Set the joint recurrence plot to fixed thresholds.

        Modifies / sets the class variables :attr:`JR` and :attr:`N`
        accordingly.

        :type threshold: tuple of number
        :arg threshold: The recurrence threshold. Give for both time series
            separately.
        """
        if self.silence_level <= 1:
            print("Calculating joint recurrence plot at fixed threshold...")

        self.embedding = self.x_embedded
        distance = self.distance_matrix(self.metric[0])
        N = distance.shape[0]
        recurrence_x = np.zeros((N, N), dtype="int8")
        recurrence_x[distance < threshold[0]] = 1
        del distance

        self.embedding = self.y_embedded
        distance = self.distance_matrix(self.metric[1])
        recurrence_y = np.zeros((N, N), dtype="int8")
        recurrence_y[distance < threshold[1]] = 1
        del distance

        if self.lag >= 0:
            self.JR = recurrence_x[:N-self.lag, :N-self.lag] * \
                recurrence_y[self.lag:N, self.lag:N]
        else:
            # self.JR = recurrence_x[-self.lag:N, -self.lag:N] * \
            #     recurrence_y[:N+self.lag, :N+self.lag]
            self.JR = recurrence_y[:N+self.lag, :N+self.lag] * \
                recurrence_x[-self.lag:N, -self.lag:N]
        self.N = N

    def set_fixed_threshold_std(self, threshold_std):
        """
        Set the joint recurrence plot to fixed thresholds in units of the
        standard deviation of the time series.

        Calculates the absolute thresholds and calls
        :meth:`set_fixed_threshold`.

        :type threshold_std: tuple of number
        :arg threshold_std: The recurrence threshold in units of the standard
            deviation of the time series. Give for both time series separately.
        """
        if self.silence_level <= 1:
            print("Calculating recurrence plot at fixed threshold "
                  "in units of time series STD...")

        threshold_x = threshold_std[0] * self.x.std()
        threshold_y = threshold_std[1] * self.y.std()
        JointRecurrencePlot.\
            set_fixed_threshold(self, (threshold_x, threshold_y))

    def set_fixed_recurrence_rate(self, recurrence_rate):
        """
        Set the joint recurrence plot to fixed recurrence rates.

        Modifies / sets the class variables :attr:`JR` and :attr:`N`
        accordingly.

        :type recurrence_rate: tuple of number
        :arg recurrence_rate: The recurrence rate. Give for both time series
            separately.
        """
        if self.silence_level <= 1:
            print("Calculating joint recurrence plot at "
                  "fixed recurrence rate...")

        self.embedding = self.x_embedded
        distance = self.distance_matrix(self.metric[0])
        N = distance.shape[0]
        threshold_x = self.\
            threshold_from_recurrence_rate(distance, recurrence_rate[0])
        recurrence_x = np.zeros((N, N), dtype="int8")
        recurrence_x[distance < threshold_x] = 1
        del distance

        self.embedding = self.y_embedded
        distance = self.distance_matrix(self.metric[1])
        threshold_y = self.\
            threshold_from_recurrence_rate(distance, recurrence_rate[1])
        recurrence_y = np.zeros((N, N), dtype="int8")
        recurrence_y[distance < threshold_y] = 1
        del distance

        if self.lag >= 0:
            self.JR = recurrence_x[:N-self.lag, :N-self.lag] * \
                recurrence_y[self.lag:N, self.lag:N]
        else:
            # self.JR = recurrence_x[-self.lag:N, -self.lag:N] * \
            #     recurrence_y[:N+self.lag, :N+self.lag]
            self.JR = recurrence_y[:N+self.lag, :N+self.lag] * \
                recurrence_x[-self.lag:N, -self.lag:N]
        self.N = N



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides class for spatio-temporal grids.
"""

from typing import Tuple
from collections.abc import Hashable
import pickle

import numpy as np

from .cache import Cached
from ._ext.types import to_cy, FIELD
from ._ext.numerics import _calculate_euclidean_distance


class Grid(Cached):
    """
    Encapsulates a spatio-temporal grid.

    The spatial grid points can be arbitrarily distributed, which is useful
    for representing station data or geodesic grids.
    """

    #
    #  Definitions of internal methods
    #

    def __init__(self, time_seq: np.ndarray, space_seq: np.ndarray,
                 silence_level: int = 0):
        """
        Initialize an instance of Grid.

        :type time_seq: 1D Numpy array [time]
        :arg time_seq: The increasing sequence of temporal sampling points.

        :type space_seq: 2D Numpy array [dim, index]
        :arg space_seq: The sequences of spatial sampling points.

        :type silence_level: number (int)
        :arg silence_level: The inverse level of verbosity of the object.
        """

        #  Set basic dictionaries
        self._grid = {"time": time_seq.astype("float32"),
                      "space": space_seq.astype("float32")}
        self._grid_size = {"time": len(time_seq),
                           "space": space_seq.shape[1]}

        self._boundaries = {"time_min": time_seq.min(),
                            "time_max": time_seq.max(),
                            "space_min": np.amin(space_seq, axis=1),
                            "space_max": np.amax(space_seq, axis=1)}

        #  Defines the number of spatial grid points / nodes at one instant
        #  of time
        self.N = self._grid_size["space"]
        """(number (int)) - The number of spatial grid points / nodes."""

        #  Set silence level
        self.silence_level = silence_level
        """(number (int)) - The inverse level of verbosity of the object."""

        #  Defines the total number of data points / grid points / samples of
        #  the corresponding data set.
        self.n_grid_points = self._grid_size["time"] * self.N
        """(number (int)) - The total number of data points / samples."""

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        # The following attributes are assumed immutable:
        #   (N, _grid)
        return ()

    def __str__(self):
        """
        Return a string representation of the Grid object.
        """
        return (f"Grid: {self._grid_size['space']} grid points, "
                f"{self._grid_size['time']} timesteps.")

    #
    #  Functions for loading and saving the Grid object
    #

    def save(self, filename):
        """
        Save the Grid object to a pickle file.

        :arg str filename: The name of the file where Grid object is stored
            (including ending).
        """
        try:
            with open(filename, 'wb') as f:
                pickle.dump(self, f)
        except IOError:
            print("An error occurred while saving Grid instance to "
                  f"pickle file {filename}")

    @staticmethod
    def Load(filename):
        """
        Return a Grid object stored in a pickle file.

        :arg str filename: The name of the file where Grid object is stored
            (including ending).
        :rtype: Grid object
        :return: :class:`Grid` instance.
        """
        with open(filename, 'rb') as f:
            grid = pickle.load(f)

        return grid

    #
    #  Alternative constructors and Grid generation methods
    #

    @staticmethod
    def SmallTestGrid():
        """
        Return test grid of 6 spatial grid points with 10 temporal sampling
        points each.

        :rtype: Grid instance
        :return: a Grid instance for testing purposes.
        """
        return Grid(time_seq=np.arange(10),
                    space_seq=np.array([[0, 5, 10, 15, 20, 25],
                                        [2.5, 5., 7.5, 10., 12.5, 15.]]),
                    silence_level=2)

    @staticmethod
    def RegularGrid(time_seq, space_grid, silence_level=0):
        """
        Initialize an instance of a regular grid.

        **Examples:**

        >>> Grid.RegularGrid(
        ...      time_seq=np.arange(2),
                 space_grid=[np.array([0.,5.]), np.array([1.,2.])],
                 silence_level=2).sequence(0)
        array([ 0.,  0.,  5.,  5.], dtype=float32)
        >>> Grid.RegularGrid(
        ...     time_seq=np.arange(2),
                space_grid=[np.array([0.,5.]), np.array([1.,2.])],
                silence_level=2).sequence(1)
        array([ 1.,  2.,  1.,  2.], dtype=float32)

        :type time_seq: 1D Numpy array [time]
        :arg time_seq: The increasing sequence of temporal sampling points.

        :type space_grids: list of 1D Numpy arrays [dim, n]
        :arg space_grids: The spatial grid.

        :type silence_level: number (int)
        :arg silence_level: The inverse level of verbosity of the object.

        :rtype: Grid object
        :return: :class:`Grid` instance.
        """
        #  Generate sequences of positions for all nodes
        space_seq = Grid.coord_sequence_from_rect_grid(space_grid)

        #  Return instance of Grid
        return Grid(time_seq, space_seq, silence_level)

    #
    #  Definitions of grid related functions
    #

    @staticmethod
    def coord_sequence_from_rect_grid(space_grid):
        """
        Return the sequences of coordinates for a regular and
        rectangular grid.

        **Example:**

        >>> Grid.coord_sequence_from_rect_grid(
        ...     space_grid=[np.array([0.,5.]), np.array([1.,2.])]
        [array([ 0.,  0.,  5.,  5.]), array([ 1.,  2.,  1.,  2.])]

        :type space_grid: list of 1D Numpy arrays [dim, n]
        :arg space_grid: The grid's sampling points.

        :rtype: list of 1D Numpy arrays [index]
        :return: the coordinates of all nodes in the grid.
        """
        space_seq = np.meshgrid(*space_grid)
        space_seq = np.array([dim_seq.flatten('F') for dim_seq in space_seq])

        return space_seq

    def sequence(self, dimension):
        """
        Return the positional sequence for all nodes for the specified
        dimension.

        **Example:**

        >>> Grid.SmallTestGrid().sequence(0)
        array([  0.,   5.,  10.,  15.,  20.,  25.], dtype=float32)

        :type dimension: integer
        :arg dimension: The number of the dimension

        :rtype: 1D Numpy array [index]
        :return: the sequence of positions in the specified dimension for all
            nodes.
        """
        return self._grid["space"][dimension]

    def node_number(self, x):
        """
        Return the index of the closest node given euclidean coordinates.

        **Example:**

        >>> Grid.SmallTestGrid().node_number(x=(14., 9.))
        3

        :type x: number (float)
        :arg x: The x coordinate.

        :type y: number (float)
        :arg y: The y coordinate.

        :rtype: number (int)
        :return: the closest node's index.
        """
        x = np.array(x)

        # Get the differences of the coordinates of all nodes to the given
        # coordinates
        diff = self._grid["space"].T - x

        # Get sequences of cosLat, sinLat, cosLon and sinLon for all nodes
        dist = np.sqrt(np.sum(diff**2, axis=1))

        #  Get index of closest node
        n_node = dist.argmin()

        return n_node

    def node_coordinates(self, index):
        """
        Return the position of node ``index``.

        **Example:**

        >>> Grid.SmallTestGrid().node_coordinates(3)
        [15.0, 10.0]

        :type index: number (int)
        :arg index: The node index as used in node sequences.

        :rtype: tuple of number (float)
        :return: the node's coordinates.
        """
        return tuple(self._grid["space"][:, index])

    def distance(self):
        """
        Calculate and return the standard distance matrix of the corresponding
        grid type

        :rtype: 2D Numpy array [index, index]
        :return: the distance matrix.
        """
        return self.euclidean_distance()

    @Cached.method()
    def euclidean_distance(self):
        """
        Return the euclidean distance matrix between grid points.

        **Example:**

        >>> Grid.SmallTestGrid().euclidean_distance().round(2)
        [[ 0.    5.59 11.18 16.77 22.36 27.95]
         [ 5.59  0.    5.59 11.18 16.77 22.36]
         [11.18  5.59  0.    5.59 11.18 16.77]
         [16.77 11.18  5.59  0.    5.59 11.18]
         [22.36 16.77 11.18  5.59  0.    5.59]
         [27.95 22.36 16.77 11.18  5.59  0.  ]]

        :rtype: 2D Numpy array [index, index]
        :return: the euclidean distance matrix.
        """
        #  Get number of nodes
        N_nodes = self.N

        #  Get sequences of coordinates
        sequences = to_cy(self._grid["space"], FIELD)

        #  Get number of dimensions
        N_dim = sequences.shape[0]

        distance = np.zeros((N_nodes, N_nodes), dtype=FIELD)
        _calculate_euclidean_distance(sequences, distance, N_dim, N_nodes)
        return distance

    def boundaries(self):
        """
        Return the spatio-temporal grid boundaries.

        Structure of the returned dictionary:
          - self._boundaries = {"time_min": time_seq.min(),
                                "time_max": time_seq.max(),
                                "space_min": np.amax(space_seq, axis=1),
                                "space_max": np.amin(space_seq, axis=1)}

        :rtype: dictionary
        :return: the spatio-temporal grid boundaries.
        """
        return self._boundaries

    def grid(self):
        """
        Return the grid's spatio-temporal sampling points.

        Structure of the returned dictionary:
          - self._grid = {"time": time_seq.astype("float32"),
                          "space": space_seq.astype("float32")}

        **Examples:**

        >>> Grid.SmallTestGrid().grid()["space"][0]
        array([  0.,   5.,  10.,  15.,  20.,  25.], dtype=float32)
        >>> Grid.SmallTestGrid().grid()["space"][0][5]
        15.0

        :rtype: dictionary
        :return: the grid's spatio-temporal sampling points.
        """
        return self._grid

    def grid_size(self):
        """
        Return the sizes of the grid's spatial and temporal dimensions.

        Structure of the returned dictionary:
          - self._grid_size = {"time": len(time_seq),
                               "space": space_seq.shape[1]}

        **Example:**

        >>> print(Grid2D.SmallTestGrid().print_grid_size())
           space    time
               6      10

        :rtype: dictionary
        :return: the sizes of the grid's spatial and temporal dimensions.
        """
        return self._grid_size

    def print_grid_size(self):
        """
        Pretty print the sizes of the grid's spatial and temporal dimensions.
        """
        return "     space    time\n   {space:7} {time:7}".format(
            **self.grid_size())

    def geometric_distance_distribution(self, n_bins):
        """
        Return the distribution of distances between all pairs of grid points.

        **Examples:**

        >>> Grid.SmallTestGrid().geometric_distance_distribution(3)[0].round(2)
        array([0.33, 0.47, 0.2 ])
        >>> Grid.SmallTestGrid().geometric_distance_distribution(3)[1].round(2)
        array([ 0.  ,  9.32, 18.63, 27.95], dtype=float32)

        :type n_bins: number (int)
        :arg n_bins: The number of histogram bins.

        :rtype: tuple of two 1D Numpy arrays [bin]
        :return: the normalized histogram and lower bin boundaries of
            distances.
        """
        if self.silence_level <= 1:
            print("Calculating the geometric distance distribution of the "
                  "grid...")

        #  Get angular distance matrix
        D = self.distance()

        #  Determine range for link distance histograms
        max_range = D.max()
        interval = (0, max_range)

        #  Calculate geometry related factor of distributions to divide it out
        (dist, lbb) = np.histogram(a=D, bins=n_bins, range=interval)
        #  Subtract self.N from first bin because of spurious links with zero
        #  distance on the diagonal of the angular distance matrix
        dist[0] -= self.N

        #  Normalize distribution
        dist = dist.astype("float")
        dist /= dist.sum()

        return (dist, lbb)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

from libc.math cimport sqrt, floor

from datetime import datetime
import random

import numpy as np
import numpy.random as rd
cimport numpy as cnp
from numpy cimport ndarray

from ...core._ext.types import MASK, NODE, LAG, FIELD, DFIELD
from ...core._ext.types cimport \
    ADJ_t, MASK_t, NODE_t, DEGREE_t, LAG_t, FIELD_t, DFIELD_t

cdef extern from "src_numerics.c":
    void _test_pearson_correlation_fast(double *original_data,
        double *surrogates, float *correlation, int n_time, int N, double norm)
    void _test_mutual_information_fast(int N, int n_time, int n_bins,
        double scaling, double range_min, double *original_data,
        double *surrogates, int *symbolic_original, int *symbolic_surrogates,
        int *hist_original, int *hist_surrogates, int * hist2d, float *mi)


# cross_recurrence_plot =======================================================


def _manhattan_distance_matrix_crp(
    int ntime_x, int ntime_y, int dim,
    ndarray[DFIELD_t, ndim=2, mode='c'] x_embedded not None,
    ndarray[DFIELD_t, ndim=2, mode='c'] y_embedded not None):

    cdef:
        int j, k, l
        DFIELD_t sum
        ndarray[DFIELD_t, ndim=2, mode='c'] distance = \
            np.zeros((ntime_x, ntime_y), dtype=DFIELD)

    for j in range(ntime_x):
        for k in range(ntime_y):
            sum = 0
            for l in range(dim):
                diff = abs(x_embedded[j,l] - y_embedded[k,l])
                sum += diff
            distance[j, k] = sum
    return distance


def _euclidean_distance_matrix_crp(
    int ntime_x, int ntime_y, int dim,
    ndarray[DFIELD_t, ndim=2, mode='c'] x_embedded not None,
    ndarray[DFIELD_t, ndim=2, mode='c'] y_embedded not None):

    cdef:
        int j, k, l
        DFIELD_t sum, diff
        ndarray[DFIELD_t, ndim=2, mode='c'] distance = \
            np.zeros((ntime_x, ntime_y), dtype=DFIELD)

    for j in range(ntime_x):
        for k in range(ntime_y):
            sum = 0
            for l in range(dim):
                diff = abs(x_embedded[j,l] - y_embedded[k,l])
                sum += diff * diff
            distance[j, k] = sqrt(sum)
    return distance


def _supremum_distance_matrix_crp(
    int ntime_x, int ntime_y, int dim,
    ndarray[DFIELD_t, ndim=2, mode='c'] x_embedded not None,
    ndarray[DFIELD_t, ndim=2, mode='c'] y_embedded not None):

    cdef:
        int j, k, l
        DFIELD_t temp_diff, diff
        ndarray[DFIELD_t, ndim=2, mode='c'] distance = \
            np.zeros((ntime_x, ntime_y), dtype=DFIELD)

    for j in range(ntime_x):
        for k in range(ntime_y):
            diff = 0
            for l in range(dim):
                temp_diff = abs(x_embedded[j,l] - y_embedded[k,l])
                if temp_diff > diff:
                    diff = temp_diff
            distance[j, k] = diff
    return distance


# surrogates ==================================================================


def _embed_time_series_array(
    int n, int n_time, int dimension, int delay,
    ndarray[DFIELD_t, ndim=2] time_series_array,
    ndarray[DFIELD_t, ndim=3] embedding):

    cdef:
        int i, j, k, max_delay, len_embedded, index
        int N = n, D = dimension

    # Calculate the maximum delay
    max_delay = (dimension - 1) * delay
    # Calculate the length of the embedded time series
    len_embedded = n_time - max_delay

    for i in range(N):
        for j in range(D):
            index = j*delay
            for k in range(len_embedded):
                embedding[i, k, j] = time_series_array[i, index]
                index += 1


def _recurrence_plot(
    int n_time, int dimension, float threshold,
    ndarray[DFIELD_t, ndim=2] embedding,
    ndarray[ADJ_t, ndim=2] R):

    cdef:
        int j, k, l, T = n_time, D = dimension
        DFIELD_t diff

    for j in range(T):
        # Ignore the main diagonal, since every sample is neighbor of itself
        for k in range(j):
            for l in range(D):
                # Use supremum norm
                diff = embedding[j, l] - embedding[k, l]

                if abs(diff) > threshold:
                    # j and k are not neighbors
                    R[j, k] = R[k, j] = 0

                    # Leave the loop
                    break


def _twins_s(
    int N, int n_time, int dimension, float threshold, int min_dist,
    ndarray[DFIELD_t, ndim=3] embedding_array,
    ndarray[ADJ_t, ndim=2] R, ndarray[DEGREE_t, ndim=1] nR,
    twins):

    cdef:
        int i, j, k, l
        DFIELD_t diff
        object twins_i, twins_ij, twins_ik

    for i in range(N):
        # Initialize the recurrence matrix R and nR
        for j in range(n_time):
            for k in range(j+1):
                R[j, k] = R[k, j] = 1
            nR[j] = <DEGREE_t> n_time

        # Calculate the recurrence matrix for time series i
        for j in range(n_time):
            # Ignore main diagonal, since every sample is neighbor of itself
            for k in range(j):
                for l in range(dimension):
                    # Use maximum norm
                    diff = embedding_array[i, j, l] - embedding_array[i, k, l]

                    if abs(diff) > threshold:
                        # j and k are not neighbors
                        R[j, k] = R[k, j] = 0

                        # Reduce neighbor count of j and k by one
                        nR[j] -= 1
                        nR[k] -= 1

                        # Leave the for loop
                        break

        # Add list for twins in time series i
        twins.append([])

        # Find all twins in the recurrence matrix
        for j in range(n_time):
            twins_i = twins[i]
            twins_i.append([])
            twins_ij = twins_i[j]

            # Respect a minimal temporal spacing between twins to avoid false
            # twins due to the higher
            # sample density in phase space along the trajectory
            for k in range(j - min_dist):
                # Continue only if both samples have the same number of
                # neighbors and more than just one neighbor (themselves)
                if nR[j] == nR[k] and nR[j] != 1:
                    l = 0

                    while R[j, l] == R[k, l]:
                        l += 1

                        # If l is equal to the length of the time series at
                        # this point, j and k are twins
                        if l == n_time:
                            # Add the twins to the twin list
                            twins_ik = twins_i[k]

                            twins_ij.append(k)
                            twins_ik.append(j)

                            # Leave the while loop
                            break


def _twin_surrogates_s(int n_surrogates, int N, twins,
                       ndarray[DFIELD_t, ndim=2] original_data):
    cdef:
        int i, j, k, new_k, n_twins, rand
        object twins_i, twins_ik
        ndarray[DFIELD_t, ndim=2] surrogates = np.empty(
            (n_surrogates, N), dtype=DFIELD)

    for i in range(n_surrogates):
        # Get the twin list for time series i
        twins_i = twins[i]
        # Randomly choose a starting point in the original trajectory
        k = int(floor(random.random() * N))

        j = 0
        while j < N:
            # Assign state vector of surrogate trajectory
            surrogates[i,j] = original_data[i,k]
            # Get the list of twins of state vector k in the original time
            # series
            twins_ik = twins_i[k]

            # Get the number of twins of k
            n_twins = int(len(twins_ik))
            # If k has no twins, go to the next sample k+1, If k has twins at
            # m, choose among m+1 and k+1 with equal probability
            if n_twins == 0:
                k += 1
            else:
                # Generate a random integer between 0 and n_twins
                rand = int(floor(random.random() * (n_twins + 1)))

                # If rand = n_twings go to smple k+1, otherwise jump to the
                # future of one of the twins
                if rand == n_twins:
                    k += 1
                else:
                    k = twins_ik[rand]
                    k += 1

            # If the new k >= n_time, choose a new random starting point in the
            # original time series
            if k >= N:
                while True:
                    new_k = int(floor(random.random() * N))
                    if new_k != k:
                        break
                k = new_k

            j += 1

    return surrogates


def _test_pearson_correlation(
    ndarray[DFIELD_t, ndim=2, mode='c'] original_data not None,
    ndarray[DFIELD_t, ndim=2, mode='c'] surrogates not None,
    int N, int n_time):

    cdef:
        DFIELD_t norm = 1.0 / float(n_time)
        #  Initialize Pearson correlation matrix
        ndarray[FIELD_t, ndim=2, mode='c'] correlation = np.zeros(
            (N, N), dtype=FIELD)

    _test_pearson_correlation_fast(
        <DFIELD_t*> cnp.PyArray_DATA(original_data),
        <DFIELD_t*> cnp.PyArray_DATA(surrogates),
        <FIELD_t*> cnp.PyArray_DATA(correlation),
        n_time, N, norm)

    return correlation


def _test_mutual_information(
    ndarray[DFIELD_t, ndim=2, mode='c'] original_data not None,
    ndarray[DFIELD_t, ndim=2, mode='c'] surrogates not None,
    int N, int n_time, int n_bins):

    cdef:
        #  Get common range for all histograms
        DFIELD_t range_min = np.min((original_data.min(), surrogates.min()))
        DFIELD_t range_max = np.max((original_data.max(), surrogates.max()))
        #  Rescale all time series to the interval [0,1], using the maximum
        #  range of the whole dataset
        DFIELD_t scaling = 1. / (range_max - range_min)
        #  Create arrays to hold symbolic trajectories
        ndarray[NODE_t, ndim=2, mode='c'] symbolic_original = \
            np.empty((N, n_time), dtype=NODE)
        ndarray[NODE_t, ndim=2, mode='c'] symbolic_surrogates = \
            np.empty((N, n_time), dtype=NODE)
        #  Initialize array to hold 1d-histograms of individual time series
        ndarray[NODE_t, ndim=2, mode='c'] hist_original = \
            np.zeros((N, n_bins), dtype=NODE)
        ndarray[NODE_t, ndim=2, mode='c'] hist_surrogates = \
            np.zeros((N, n_bins), dtype=NODE)
        #  Initialize array to hold 2d-histogram for one pair of time series
        ndarray[NODE_t, ndim=2, mode='c'] hist2d = \
            np.zeros((n_bins, n_bins), dtype=NODE)
        #  Initialize mutual information array
        ndarray[FIELD_t, ndim=2, mode='c'] mi = np.zeros((N, N), dtype=FIELD)

    _test_mutual_information_fast(
            N, n_time, n_bins, scaling, range_min,
            <DFIELD_t*> cnp.PyArray_DATA(original_data),
            <DFIELD_t*> cnp.PyArray_DATA(surrogates),
            <int*> cnp.PyArray_DATA(symbolic_original),
            <int*> cnp.PyArray_DATA(symbolic_surrogates),
            <int*> cnp.PyArray_DATA(hist_original),
            <int*> cnp.PyArray_DATA(hist_surrogates),
            <int*> cnp.PyArray_DATA(hist2d),
            <FIELD_t*> cnp.PyArray_DATA(mi))

    return mi


# recurrence plot =============================================================


def _embed_time_series(
    int n_time, int dim, int tau,
    ndarray[FIELD_t, ndim=1] time_series,
    ndarray[FIELD_t, ndim=2] embedding):

    cdef:
        int j, max_delay, index, D = dim
        int k, len_embedded

    # Calculate the maximum delay
    max_delay = (dim - 1) * tau
    # Calculate the length of the embedded time series
    len_embedded = n_time - max_delay

    for j in range(D):
        index = j * tau
        for k in range(len_embedded):
            embedding[k, j] = time_series[index]
            index += 1


def _manhattan_distance_matrix_rp(
    int n_time, int dim, ndarray[DFIELD_t, ndim=2] embedding):

    cdef:
        int j, k, l, T = n_time, D = dim
        DFIELD_t sum
        ndarray[DFIELD_t, ndim=2, mode='c'] distance = \
            np.zeros((n_time, n_time), dtype=DFIELD)

    for j in range(T):
        # Ignore the main diagonal, since every samle is neighbor of itself
        for k in range(j):
            sum = 0
            for l in range(D):
                # use manhattan norm
                sum += abs(embedding[j, l] - embedding[k, l])
            distance[j, k] = distance[k, j] = sum
    return distance


def _euclidean_distance_matrix_rp(
    int n_time, int dim, ndarray[DFIELD_t, ndim=2] embedding):

    cdef:
        int j, k, l, T = n_time, D = dim
        DFIELD_t sum, diff
        ndarray[DFIELD_t, ndim=2, mode='c'] distance = \
            np.zeros((n_time, n_time), dtype=DFIELD)

    for j in range(T):
        # Ignore the main diagonal, since every sample is neighbor of itself
        for k in range(j):
            sum = 0
            for l in range(D):
                # Use euclidean norm
                diff = abs(embedding[j, l] - embedding[k, l])
                sum += diff * diff
            distance[j, k] = distance[k, j] = sqrt(sum)
    return distance


def _supremum_distance_matrix_rp(
    int n_time, int dim, ndarray[DFIELD_t, ndim=2] embedding):

    cdef:
        int j, k, l, T = n_time, D = dim
        DFIELD_t temp_diff, diff
        ndarray[DFIELD_t, ndim=2, mode='c'] distance = \
            np.zeros((n_time, n_time), dtype=DFIELD)

    for j in range(T):
        # Ignore the main diagonal, since every sample is neighbor of itself
        for k in range(j):
            diff = 0
            for l in range(D):
                # Use supremum norm
                temp_diff = abs(embedding[j, l] - embedding[k, l])
                if temp_diff > diff:
                    diff = temp_diff
            distance[j, k] = distance[k, j] = diff
    return distance


def _set_adaptive_neighborhood_size(
    int n_time, int adaptive_neighborhood_size,
    ndarray[NODE_t, ndim=2] sorted_neighbors,
    ndarray[NODE_t, ndim=1] order,
    ndarray[LAG_t, ndim=2] recurrence):

    cdef:
        int i, j, k, l

    for i in range(adaptive_neighborhood_size):
        for j in range(n_time):
            # Get the node index to be processed
            l = order[j]

            # Start with k = i + 1, since state vectors closer than the (i+1)th
            # nearest neighbor are already connected to j at this stage
            k = i + 1
            while recurrence[l, sorted_neighbors[l, k]] == 1 and k < n_time:
                k += 1
            # add a "new" nearest neighbor of l to the recurrence plot
            recurrence[l, sorted_neighbors[l, k]] = \
                recurrence[sorted_neighbors[l, k], l] = 1


def _bootstrap_distance_matrix_manhattan(
    int n_time, int dim, ndarray[DFIELD_t, ndim=2] embedding,
    ndarray[DFIELD_t, ndim=1] distances, int M):

    cdef:
        int i, l
        ndarray[NODE_t, ndim=2] jk = rd.randint(n_time, size=(2,M))
        double sum

    for i in range(M):
        #Compute their distance
        sum = 0
        for l in range(dim):
            # Use Manhattan norm
            sum += abs(embedding[jk[0, i], l] - embedding[jk[1, i], l])

        distances[i] = sum


def _bootstrap_distance_matrix_euclidean(
    int n_time, int dim, ndarray[DFIELD_t, ndim=2] embedding,
    ndarray[DFIELD_t, ndim=1] distances, int M):

    cdef:
        int i, l
        ndarray[NODE_t, ndim=2] jk = rd.randint(n_time, size=(2,M))
        double sum, diff

    for i in range(M):
        #Compute their distance
        sum = 0
        for l in range(dim):
            # Use Manhattan norm
            diff = abs(embedding[jk[0, i], l] - embedding[jk[1, i], l])
            sum += diff * diff

        distances[i] = sqrt(sum)


def _bootstrap_distance_matrix_supremum(
    int n_time, int dim, ndarray[DFIELD_t, ndim=2] embedding,
    ndarray[DFIELD_t, ndim=1] distances, int M):

    cdef:
        int i, l
        ndarray[NODE_t, ndim=2] jk = rd.randint(n_time, size=(2,M))
        double temp_diff, diff

    for i in range(M):
        #Compute their distance
        diff = 0
        for l in range(dim):
            # Use supremum norm
            temp_diff = abs(embedding[jk[0, i], l] - embedding[jk[1, i], l])

            if temp_diff > diff:
                diff = temp_diff

            distances[i] = diff


def _rejection_sampling(
    ndarray[DFIELD_t, ndim=1] dist,
    ndarray[NODE_t, ndim=1] resampled_dist, int N, int M):

    cdef:
        int i = 0, x

    while i < M:
        x = int(floor(random.random() * N))

        if (random.random() < dist[x]):
            resampled_dist[x] += 1
            i += 1


def _twins_r(
    int min_dist, int N, ndarray[LAG_t, ndim=2] R,
    ndarray[NODE_t, ndim=1] nR, twins):

    cdef:
        int j, k, l
        object twins_j, twins_k

    twins.append([])

    for j in range(N):
        twins.append([])
        twins_j = twins[j]

        # Respect a minimal temporal spacing between twins to avoid false
        # twins du to th higher sample density in phase space along the
        # trajectory
        for k in range(j - min_dist):
            # Continue only if both samples have the same number of
            # neighbors and more than just one neighbor (themselves)
            if nR[j] == nR[k] and nR[j] != 1:
                l = 0

                while R[j, l] == R[k, l]:
                    l = l + 1

                    # If l is equal to the length of the time series at
                    # this point, j and k are twins
                    if l == N:
                        # And the twins to the twin list
                        twins_k = twins[k]

                        twins_j.append(k)
                        twins_k.append(j)
                        break


def _twin_surrogates_r(int n_surrogates, int N, int dim, twins,
                       ndarray[DFIELD_t, ndim=2] embedding):
    cdef:
        int i, j, k, new_k, n_twins, rand
        object twins_k
        ndarray[DFIELD_t, ndim=2] surrogates = np.empty(
            (n_surrogates, N, dim), dtype=DFIELD)

    # Initialize random number generator
    random.seed(datetime.now())

    for i in range(n_surrogates):
        # Randomly choose a starting point in the original trajectory
        k = int(floor(random.random() * N))

        j = 0
        while j < N:
            # Assign state vector of surrogate trajectory
            surrogates[i, j, :] = embedding[k, :]
            # Get the list of twins of state vector k in the original time
            # series
            twins_k = twins[k]

            # Get the number of twins of k
            n_twins = int(len(twins_k))
            # If k has no twins, go to the next sample k+1, If k has twins at
            # m, choose among m+1 and k+1 with equal probability
            if n_twins == 0:
                k += 1
            else:
                # Generate a random integer between 0 and n_twins
                rand = int(floor(random.random() * (n_twins + 1)))

                # If rand = n_twings go to smple k+1, otherwise jump to the
                # future of one of the twins
                if rand == n_twins:
                    k += 1
                else:
                    k = twins_k[rand]
                    k += 1

            # If the new k >= n_time, choose a new random starting point in the
            # original time series
            if k >= N:
                while True:
                    new_k = int(floor(random.random() * N))
                    if new_k != k:
                        break
                k = new_k

            j += 1

    return surrogates



# recurrence line distributions ===============================================


# parameters for `_line_dist()`
ctypedef int     (*line_type_i2J) (int, int)
ctypedef int     (*line_type_ij2I)(int, int, int)
ctypedef DFIELD_t (*metric_type)(int, int, int, DFIELD_t[:,:])

cdef:
    inline int i2J_vertline(int i, int N): return N
    inline int i2J_diagline(int i, int N): return i+1
    inline int ij2I_vertline(int i, int j, int N): return i
    inline int ij2I_diagline(int i, int j, int N): return N - i + j
    metric_type metric_null = NULL
    inline DFIELD_t metric_supremum(int I, int j, int dim, DFIELD_t[:,:] E):
        cdef:
            int l
            DFIELD_t diff = 0, tmp_diff
        for l in range(dim):
            tmp_diff = abs(E[I, l] - E[j, l])
            if tmp_diff > diff:
                diff = tmp_diff
        return diff


cdef void _line_dist(
    int n_time, ndarray[NODE_t, ndim=1] hist,
    ndarray[LAG_t, ndim=2] R, ndarray[DFIELD_t, ndim=2] E, float eps, int dim,
    metric_type metric, bint black,
    ndarray[MASK_t, ndim=1, cast=True] M, bint missing_values,
    line_type_i2J i2J, line_type_ij2I ij2I, bint skip_main):
    """
    Recurrence line distributions, parametrised by the following arguments:

      - `R` | `E (dim > 0)`: recurrence computation (cached vs. raw embedding)
      - `metric`: embedding metric (currently only supremum)
      - `black`: RQA colour (black vs. white)
      - `M (missing_values == 1)`: missing input values (ignore vs. account)
      - `line_type_*`, `skip_main`: line type (vertical vs. diagonal)
    """

    cdef:
        int i, I, j, k = 0, N = n_time
        FIELD_t d
        bint line, missing_flag = False

    if skip_main:
        # exclude main diagonal by skipping last outer loop iteration
        N -= 1
    for i in range(N):
        for j in range(i2J(i, N)):
            I = ij2I(i, j, N)

            if dim == 0:
                line = R[I, j] == black
            else:
                # compute distance between embedding vectors
                d = metric(I, j, dim, E)
                line = (d < eps) == black

            if missing_values:
                # check if current point in RP is a missing value
                if M[I] or M[j]:
                    missing_flag = True
                    k = 0
                # or if previous point was one, reset flag if not within line
                elif missing_flag and not line:
                    missing_flag = False

                # only count line if it does not contain,
                # directly follow or is followed by a missing value
                if missing_flag:
                    continue

            if line:
                # if within line, increment length
                k += 1
            elif k != 0:
                # if end of line, count line and reset length
                hist[k-1] += 1
                k = 0

        if k != 0 and not missing_flag:
            # at end of subspace, count the last uncounted line and reset length
            hist[k-1] += 1
            k = 0
        missing_flag = False


def _vertline_dist(
        int n_time, ndarray[NODE_t, ndim=1] hist, ndarray[LAG_t, ndim=2] R):
    cdef:
        ndarray[DFIELD_t, ndim=2] E_null = np.array([[]], dtype=DFIELD)
        ndarray[MASK_t, ndim=1] M_null = np.array([], dtype=MASK)
    _line_dist(
        n_time, hist, R, E_null, 0, 0, metric_null, True, M_null, False,
        i2J_vertline, ij2I_vertline, False)

def _diagline_dist(
        int n_time, ndarray[NODE_t, ndim=1] hist, ndarray[LAG_t, ndim=2] R):
    cdef:
        ndarray[DFIELD_t, ndim=2] E_null = np.array([[]], dtype=DFIELD)
        ndarray[MASK_t, ndim=1] M_null = np.array([], dtype=MASK)
    _line_dist(
        n_time, hist, R, E_null, 0, 0, metric_null, True, M_null, False,
        i2J_diagline, ij2I_diagline, True)

def _white_vertline_dist(
        int n_time, ndarray[NODE_t, ndim=1] hist, ndarray[LAG_t, ndim=2] R):
    cdef:
        ndarray[DFIELD_t, ndim=2] E_null = np.array([[]], dtype=DFIELD)
        ndarray[MASK_t, ndim=1] M_null = np.array([], dtype=MASK)
    _line_dist(
        n_time, hist, R, E_null, 0, 0, metric_null, False, M_null, False,
        i2J_vertline, ij2I_vertline, False)

def _vertline_dist_sequential(
        int n_time, ndarray[NODE_t, ndim=1] hist,
        ndarray[DFIELD_t, ndim=2] E, float eps, int dim):
    cdef:
        ndarray[LAG_t, ndim=2] null_R = np.array([[]], dtype=LAG)
        ndarray[MASK_t, ndim=1] M_null = np.array([], dtype=MASK)
    _line_dist(
        n_time, hist, null_R, E, eps, dim, metric_supremum, True, M_null, False,
        i2J_vertline, ij2I_vertline, False)

def _diagline_dist_sequential(
        int n_time, ndarray[NODE_t, ndim=1] hist,
        ndarray[DFIELD_t, ndim=2] E, float eps, int dim):
    cdef:
        ndarray[LAG_t, ndim=2] null_R = np.array([[]], dtype=LAG)
        ndarray[MASK_t, ndim=1] M_null = np.array([], dtype=MASK)
    _line_dist(
        n_time, hist, null_R, E, eps, dim, metric_supremum, True, M_null, False,
        i2J_diagline, ij2I_diagline, True)

def _vertline_dist_missingvalues(
        int n_time, ndarray[NODE_t, ndim=1] hist, ndarray[LAG_t, ndim=2] R,
        ndarray[MASK_t, ndim=1, cast=True] M):
    cdef:
        ndarray[DFIELD_t, ndim=2] E_null = np.array([[]], dtype=DFIELD)
    _line_dist(
        n_time, hist, R, E_null, 0, 0, metric_null, True, M, True,
        i2J_vertline, ij2I_vertline, False)

def _diagline_dist_missingvalues(
        int n_time, ndarray[NODE_t, ndim=1] hist, ndarray[LAG_t, ndim=2] R,
        ndarray[MASK_t, ndim=1, cast=True] M):
    cdef:
        ndarray[DFIELD_t, ndim=2] E_null = np.array([[]], dtype=DFIELD)
    _line_dist(
        n_time, hist, R, E_null, 0, 0, metric_null, True, M, True,
        i2J_diagline, ij2I_diagline, True)

def _vertline_dist_sequential_missingvalues(
        int n_time, ndarray[NODE_t, ndim=1] hist,
        ndarray[MASK_t, ndim=1, cast=True] M,
        ndarray[DFIELD_t, ndim=2] E, float eps, int dim):
    cdef:
        ndarray[LAG_t, ndim=2] null_R = np.array([[]], dtype=LAG)
    _line_dist(
        n_time, hist, null_R, E, eps, dim, metric_supremum, True, M, True,
        i2J_vertline, ij2I_vertline, False)

def _diagline_dist_sequential_missingvalues(
        int n_time, ndarray[NODE_t, ndim=1] hist,
        ndarray[MASK_t, ndim=1, cast=True] M,
        ndarray[DFIELD_t, ndim=2] E, float eps, int dim):
    cdef:
        ndarray[LAG_t, ndim=2] null_R = np.array([[]], dtype=LAG)
    _line_dist(
        n_time, hist, null_R, E, eps, dim, metric_supremum, True, M, True,
        i2J_diagline, ij2I_diagline, True)


# visibility graph =============================================================


def _visibility_relations_missingvalues(
    ndarray[FIELD_t, ndim=1] x, ndarray[FIELD_t, ndim=1] t,
    int N, ndarray[MASK_t, ndim=2] A,
    ndarray[MASK_t, ndim=1, cast=True] mv_indices):

    cdef:
        int i, j, k
        FIELD_t test

    for i in range(N-2):
        for j in range(i+2, N):
            k = i + 1

            test = (x[j] - x[i]) / (t[j] - t[i])

            while not mv_indices[k] and\
                (x[k] - x[i]) / (t[k] - t[i]) < test and k < j:
                k += 1

            if k == j:
                A[i, j] = A[j, i] = 1

    # Add trivial connections of subsequent observations in time series
    for i in range(N-1):
        if not mv_indices[i] and not mv_indices[i+1]:
            A[i, i+1] = A[i+1, i] = 1


def _visibility_relations_no_missingvalues(
    ndarray[FIELD_t, ndim=1] x, ndarray[FIELD_t, ndim=1] t,
    int N, ndarray[MASK_t, ndim=2] A):

    cdef:
        int i, j, k
        FIELD_t test

    for i in range(N-2):
        for j in range(i+2, N):
            k = i + 1

            test = (x[j] - x[i]) / (t[j] - t[i])

            while (x[k] - x[i]) / (t[k] - t[i]) < test and k < j:
                k += 1

            if k == j:
                A[i, j] = A[j, i] = 1

    # Add trivial connections of subsequent observations in time series
    for i in range(N-1):
        A[i, i+1] = A[i+1, i] = 1


def _visibility_relations_horizontal(
    ndarray[FIELD_t, ndim=1] x, int N, ndarray[MASK_t, ndim=2] A):

    cdef:
        int i, j, k
        FIELD_t minimum

    for i in range(N-2):
        for j in range(i+2, N):
            k = i + 1
            minimum = min(x[i], x[j])

            while x[k] < minimum and k < j:
                k += 1

            if k == j:
                A[i, j] = A[j, i] = 1

    # Add trivial connections of subsequent observations in time series
    for i in range(N-1):
        A[i, i+1] = A[i+1, i] = 1


def _retarded_local_clustering(
    int N, ndarray[ADJ_t, ndim=2] A,
    ndarray[DFIELD_t, ndim=1] norm,
    ndarray[DFIELD_t, ndim=1] retarded_clustering):

    cdef:
        int i, j, k
        long counter

    # Loop over all nodes
    for i in range(N):
        # Check if i has right degree larger than 1
        if norm[i] != 0:
            # Reset counter
            counter = 0

            # Loop over unique pairs of nodes in the past of i
            for j in range(i):
                for k in range(j):
                    if A[i, j] == 1 and A[j, k] == 1 and A[k, i] == 1:
                        counter += 1

            retarded_clustering[i] = counter / norm[i]


def _advanced_local_clustering(
    int N, ndarray[ADJ_t, ndim=2] A,
    ndarray[DFIELD_t, ndim=1] norm,
    ndarray[DFIELD_t, ndim=1] advanced_clustering):

    cdef:
        int i, j, k
        long counter

    # Loop over all nodes
    for i in range(N-2):
        # Check if i has right degree larger than 1
        if norm[i] != 0:
            # Reset counter
            counter = 0

            # Loop over unique pairs of nodes in the future of i
            for j in range(i+1, N):
                for k in range(i+1, j):
                    if A[i, j] == 1 and A[j, k] == 1 and A[k, i] == 1:
                        counter += 1

            advanced_clustering[i] = counter / norm[i]



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Simple tests for the funcnet CouplingAnalysis class.
"""
import numpy as np

from pyunicorn.core.data import Data
from pyunicorn.funcnet import CouplingAnalysis

from pyunicorn.core._ext.types import FIELD


def create_test_data():
    # Create test time series
    tdata = Data.SmallTestData().observable()
    # subtract means form the input data
    tdata -= np.mean(tdata, axis=1)[:, None]
    # normalize the data
    tdata /= np.sqrt(np.sum(tdata*tdata, axis=1))[:, None]
    return tdata


def test_symmetrize_by_absmax():
    # Test example
    ca = CouplingAnalysis(CouplingAnalysis.test_data())
    similarity_matrix, lag_matrix = ca.cross_correlation(tau_max=2)

    res = (similarity_matrix, lag_matrix)
    exp = (np.array([[1., 0.698, 0.7788, 0.7535],
                     [0.4848, 1., 0.4507, 0.52],
                     [0.6219, 0.5704, 1., 0.5996],
                     [0.4833, 0.5503, 0.5002, 1.]]),
           np.array([[0, 2, 1, 2], [0, 0, 0, 0],
                     [0, 2, 0, 1], [0, 2, 0, 0]]))
    assert np.allclose(res, exp, atol=1e-04)

    res = ca.symmetrize_by_absmax(similarity_matrix, lag_matrix)
    exp = (np.array([[1., 0.698, 0.7788, 0.7535],
                     [0.698, 1., 0.5704, 0.5503],
                     [0.7788, 0.5704, 1., 0.5996],
                     [0.7535, 0.5503, 0.5996, 1.]]),
           np.array([[0, 2, 1, 2], [-2, 0, -2, -2],
                     [-1, 2, 0, 1], [-2, 2, -1, 0]]))
    assert np.allclose(res, exp, atol=1e-04)

    # Random consistency test
    tdata = create_test_data()
    n_index, n_times = tdata.shape
    coup_ana = CouplingAnalysis(tdata)
    similarity_matrix = np.random.rand(n_index, n_times).astype(FIELD)
    lag_matrix = np.random.rand(n_index, n_times).astype(np.int8)
    sm_new = coup_ana.symmetrize_by_absmax(similarity_matrix, lag_matrix)[0]
    for i in range(n_index):
        for j in range(n_times):
            assert sm_new[i, j] >= similarity_matrix[i, j]


def test_cross_correlation_max():
    coup_ana = CouplingAnalysis(CouplingAnalysis.test_data())
    similarity_matrix, lag_matrix = coup_ana.cross_correlation(
        tau_max=5, lag_mode='max')
    res = (similarity_matrix, lag_matrix)
    exp = (np.array([[1., 0.7570, 0.7790, 0.7536],
                     [0.4847, 1., 0.4502, 0.5197],
                     [0.6219, 0.5844, 1., 0.5992],
                     [0.4827, 0.5509, 0.4996, 1.]]),
           np.array([[0, 4, 1, 2], [0, 0, 0, 0], [0, 3, 0, 1], [0, 2, 0, 0]]))
    assert np.allclose(res, exp, atol=1e-04)


def test_cross_correlation_all():
    coup_ana = CouplingAnalysis(CouplingAnalysis.test_data())
    res = coup_ana.cross_correlation(tau_max=1, lag_mode='all')
    exp = np.array(
        [[[1., 0.8173], [0.4849, 0.5804], [0.6214, 0.7786], [0.4831, 0.6042]],
         [[0.4849, 0.4101], [1., 0.9362], [0.4503, 0.3780], [0.5199, 0.4286]],
         [[0.6214, 0.5178], [0.4503, 0.5376], [1., 0.4962], [0.5004, 0.5996]],
         [[0.4831, 0.3762], [0.5199, 0.5404], [0.5004, 0.4092], [1., 0.4380]]])
    assert np.allclose(res, exp, atol=1e-04)


def test_mutual_information():
    coup_ana = CouplingAnalysis(CouplingAnalysis.test_data())
    similarity_matrix, lag_matrix = coup_ana.mutual_information(
        tau_max=5, knn=10, estimator='knn')
    res = (similarity_matrix, lag_matrix)
    exp = (np.array([[4.6505, 0.4387, 0.4652, 0.4126],
                     [0.147, 4.6505, 0.1065, 0.1639],
                     [0.2483, 0.2126, 4.6505, 0.2204],
                     [0.1209, 0.199, 0.1453, 4.6505]]),
           np.array([[0, 4, 1, 2],
                     [0, 0, 0, 0],
                     [0, 2, 0, 1],
                     [0, 2, 0, 0]], dtype=np.int8))
    assert np.allclose(res, exp, atol=1e-04)


def test_information_transfer():
    coup_ana = CouplingAnalysis(CouplingAnalysis.test_data())
    similarity_matrix, lag_matrix = coup_ana.information_transfer(
        tau_max=5, estimator='knn', knn=10)
    res = (similarity_matrix, lag_matrix)
    exp = (np.array([[0., 0.1544, 0.3261, 0.3047],
                     [0.0218, 0., 0.0394, 0.0976],
                     [0.0134, 0.0663, 0., 0.1502],
                     [0.0066, 0.0694, 0.0401, 0.]]),
           np.array([[0, 2, 1, 2], [5, 0, 0, 0], [5, 1, 0, 1], [5, 0, 0, 0]]))
    assert np.allclose(res, exp, atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex coupled climate networks.
"""

import numpy as np

from .coupled_climate_network import CoupledClimateNetwork


# pylint: disable=too-many-ancestors
class CoupledTsonisClimateNetwork(CoupledClimateNetwork):

    """
    Encapsulates a coupled similarity network embedded on a spherical surface.

    Particularly provides functionality to generate a complex network from the
    Pearson correlation matrix of time series from two different
    observables (temperature, pressure), vertical levels etc.

    Construct a static climate network following Tsonis et al. from the
    Pearson correlation matrix at zero lag [Tsonis2004]_.

    Hence, coupled Tsonis climate networks are undirected due to the symmetry
    of the correlation matrix.

    The idea of coupled climate networks is based on the concept of coupled
    patterns, for a review refer to [Bretherton1992]_.

    .. note::
       The two observables (layers) need to have the same time grid \
       (temporal sampling points).
    """
    #
    #  Definitions of internal methods
    #

    def __init__(self, data_1, data_2, threshold=None, link_density=None,
                 non_local=False, node_weight_type="surface",
                 selected_months=None, silence_level=0):
        """
        Initialize an instance of :class:`CoupledTsonisClimateNetwork`.

        .. note::
           Either threshold **OR** link_density have to be given!

        Possible choices for ``node_weight_type``:
          - ``None`` (constant unit weights)
          - ``"surface"`` (cos lat)
          - ``"irrigation"`` (cos**2 lat)

        :type data_1: :class:`.ClimateData`
        :arg data_1: The climate data for the first layer.
        :type data_2: :class:`.ClimateData`
        :arg data_2: The climate data for the second layer.
        :arg float threshold: The threshold of similarity measure, above which
            two nodes are linked in the network.
        :arg float link_density: The networks's desired link density.
        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        :arg str node_weight_type: The type of geographical node weight to be
            used.
        :arg [int ]selected_months: The months for which to calculate the
            correlation matrix. The full time series are used for default value
            None.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        self.silence_level = silence_level
        """(string) - The inverse level of verbosity of the object."""

        n_time_1 = data_1.grid.grid_size()["time"]
        n_time_2 = data_2.grid.grid_size()["time"]

        #  The time series from both observables have to have the same length
        if n_time_1 == n_time_2:
            #  Prepare
            if selected_months is None:
                anomaly_1 = data_1.anomaly()
                anomaly_2 = data_2.anomaly()
            else:
                anomaly_1 = data_1.anomaly_selected_months(selected_months)
                anomaly_2 = data_2.anomaly_selected_months(selected_months)

            correlation = self._calculate_correlation(anomaly_1=anomaly_1,
                                                      anomaly_2=anomaly_2)

            #  Call the constructor of the parent class ClimateNetwork
            CoupledClimateNetwork.__init__(self, grid_1=data_1.grid,
                                           grid_2=data_2.grid,
                                           similarity_measure=correlation,
                                           threshold=threshold,
                                           link_density=link_density,
                                           non_local=non_local,
                                           directed=False,
                                           node_weight_type=node_weight_type,
                                           silence_level=silence_level)
        else:
            print("\nThe two observables (layers) have to have the same "
                  "number of temporal sampling points!\n")

    def __str__(self):
        """
        Return a string representation of CoupledClimateNetwork object.
        """
        return ('CoupledTsonisClimateNetwork:\n'
                f'{CoupledClimateNetwork.__str__(self)}')

    #
    #  Defines methods to calculate the correlation matrix
    #

    def _calculate_correlation(self, anomaly_1, anomaly_2):
        """
        Return the correlation matrix at zero lag.

        :type anomaly_1: 2D Numpy array (time, index_1)
        :arg anomaly_1: the first set of anomaly time series from which to
                        calculate the correlation matrix at zero lag.

        :type anomaly_2: 2D Numpy array (time, index_2)
        :arg anomaly_2: the second set of anomaly time series from which to
                        calculate the correlation matrix at zero lag.

        :rtype: 2D Numpy array (index, index)
        :return: the correlation matrix at zero lag.
        """
        if self.silence_level <= 1:
            print("Calculating correlation matrix at zero lag from anomaly "
                  "values...")

        anomaly = np.concatenate((anomaly_1, anomaly_2), axis=1)

        #  Cast to float32 type to save memory since correlation coefficients
        #  are not needed in high floating point precision.
        correlation = np.corrcoef(anomaly.transpose()).astype("float32")

        return correlation

    def calculate_similarity_measure(self, anomaly_1, anomaly_2):
        """
        Encapsulate the calculation of the correlation matrix at zero lag.

        :type anomaly_1: 2D Numpy array (time, index_1)
        :arg anomaly_1: the first set of anomaly time series from which to
                        calculate the correlation matrix at zero lag.

        :type anomaly_2: 2D Numpy array (time, index_2)
        :arg anomaly_2: the second set of anomaly time series from which to
                        calculate the correlation matrix at zero lag.

        :rtype: 2D Numpy array (index, index)
        :return: the correlation matrix at zero lag.
        """
        return self._calculate_correlation(anomaly_1, anomaly_2)

    def correlation(self):
        """
        Return the coupled correlation matrix at zero lag.

        :rtype: 2D Numpy array (index, index)
        :return: the correlation matrix at zero lag.
        """
        return self.similarity_measure()



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex coupled climate networks.
"""

import numpy as np

from ..core import InteractingNetworks, GeoNetwork, GeoGrid
from .climate_network import ClimateNetwork


class CoupledClimateNetwork(InteractingNetworks, ClimateNetwork):
    """
    Encapsulates a coupled similarity network embedded on a spherical surface.

    Particularly provides functionality to generate a complex network from the
    matrix of a similarity measure of time series from two different
    observables (temperature, pressure), vertical levels etc.

    So far, most methods only give meaningful results for undirected networks!

    The idea of coupled climate networks is based on the concept of coupled
    patterns, for a review refer to [Bretherton1992]_.

    .. note::
       The two observables (layers) need to have the same time grid \
       (temporal sampling points).
    """
    #
    #  Definitions of internal methods
    #

    def __init__(self, grid_1, grid_2, similarity_measure, threshold=None,
                 link_density=None, non_local=False, directed=False,
                 node_weight_type="surface", silence_level=0):
        """
        Initialize an instance of CoupledClimateNetwork.

        .. note::
           Either threshold **OR** link_density have to be given!

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos**2 lat)

        :type grid_1: :class:`.GeoGrid`
        :arg  grid_1: The GeoGrid object describing the first layer's spatial
            embedding.
        :type grid_2: :class:`.GeoGrid`
        :arg  grid_2: The GeoGrid object describing the second layer's spatial
            embedding.
        :type similarity_measure: 2D array [index, index]
        :arg similarity_measure: The similarity measure for all pairs of nodes.
        :arg float threshold: The threshold of similarity measure, above which
            two nodes are linked in the network.
        :arg float link_density: The networks's desired link density.
        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        :arg bool directed: Determines, whether the network is treated as
            directed.
        :arg strnode_weight_type: The type of geographical node weight to be
            used.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        #  Store single grids
        self.grid_1 = grid_1
        """(Grid) - The GeoGrid object describing the first layer's spatial
                    embedding."""
        self.grid_2 = grid_2
        """(Grid) - The GeoGrid object describing the second layer's spatial
                    embedding."""

        #  Construct grid object describing both layers
        time_1 = grid_1.grid()["time"]
        lat_1 = grid_1.grid()["lat"]
        lon_1 = grid_1.grid()["lon"]

        time_2 = grid_2.grid()["time"]
        lat_2 = grid_2.grid()["lat"]
        lon_2 = grid_2.grid()["lon"]

        if len(time_1) == len(time_2):
            grid = GeoGrid(time_1, np.concatenate((lat_1, lat_2)),
                           np.concatenate((lon_1, lon_2)))
            #  Set total number of nodes
            self.N = grid.N
            """(number (int)) - The total number of nodes in both layers."""

            #  Set number of nodes for both layers
            self.N_1 = len(lat_1)
            """(number (int)) - The number of nodes in the first layer."""
            self.N_2 = len(lat_2)
            """(number (int)) - The number of nodes in the second layer."""

            #  Create lists of node indices for both layers
            self.nodes_1 = list(range(self.N_1))
            """(list (int)) - List of node indices for first layer"""
            self.nodes_2 = list(range(self.N_1, self.N))
            """(list (int)) - List of node indices for second layer"""

            #  Call the constructor of the parent class ClimateNetwork
            ClimateNetwork.__init__(self, grid=grid,
                                    similarity_measure=similarity_measure,
                                    threshold=threshold,
                                    link_density=link_density,
                                    non_local=non_local,
                                    directed=directed,
                                    node_weight_type=node_weight_type,
                                    silence_level=silence_level)
            InteractingNetworks.__init__(self, self.adjacency)
        else:
            print("The two observables (layers) have to have the same number "
                  "of temporal sampling points!")

    def __str__(self):
        """
        Return a string representation of CoupledClimateNetwork object.
        """
        return (f'CoupledClimateNetwork:\n{ClimateNetwork.__str__(self)}\n'
                f'N1: {self.N_1}\nN2: self.N_2')

    #
    #  Define methods for handling the coupled network
    #

    def network_1(self):
        """
        Return network consisting of layer 1 nodes and their internal links.

        This can be used to conveniently analyze the layer 1 separately, e.g.,
        for calculation network measures solely for layer 1.

        :rtype: GeoNetwork
        :return: the network consisting of layer 1 nodes and their internal
                 links.
        """
        return GeoNetwork(adjacency=self.adjacency_1(), grid=self.grid_1,
                          directed=self.directed,
                          node_weight_type=self.node_weight_type,
                          silence_level=self.silence_level)

    def network_2(self):
        """
        Return network consisting of layer 2 nodes and their internal links.

        This can be used to conveniently analyze the layer 2 separately, e.g.,
        for calculation network measures solely for layer 2.

        :rtype: GeoNetwork
        :return: the network consisting of layer 2 nodes and their internal
                 links.
        """
        return GeoNetwork(adjacency=self.adjacency_2(), grid=self.grid_2,
                          directed=self.directed,
                          node_weight_type=self.node_weight_type,
                          silence_level=self.silence_level)

    def similarity_measure_1(self):
        """
        Return internal similarity measure matrix of first layer.

        :rtype: 2D Numpy array [index_1, index_1]
        :return: the internal similarity measure matrix of first layer.
        """
        return self.similarity_measure()[:self.N_1, :self.N_1]

    def similarity_measure_2(self):
        """
        Return internal similarity measure matrix of second layer.

        :rtype: 2D Numpy array [index_2, index_2]
        :return: the internal similarity measure matrix of first layer.
        """
        return self.similarity_measure()[self.N_1:, self.N_1:]

    def cross_similarity_measure(self):
        """
        Return cross similarity measure matrix.

        .. note::
           Cross similarity measure matrix is NEITHER square NOR \
           symmetric in general!

        :rtype: 2D Numpy array [index_1, index_2]
        :return: the cross similarity measure matrix.
        """
        return self.similarity_measure()[:self.N_1, self.N_1:]

    def adjacency_1(self):
        """
        Return internal adjacency matrix of first layer.

        :rtype: 2D Numpy array [index_1, index_1]
        :return: the internal adjacency matrix of first layer.
        """
        return self.internal_adjacency(self.nodes_1)

    def adjacency_2(self):
        """
        Return internal adjacency matrix of second layer.

        :rtype: 2D Numpy array [index_2, index_2]
        :return: the internal adjacency matrix of second layer.
        """
        return self.internal_adjacency(self.nodes_2)

    def cross_layer_adjacency(self):
        """
        Return cross adjacency matrix of the coupled network.

        The cross adjacency matrix entry :math:`CA_{ij} = 1` describes that
        node :math:`i` in the first layer is linked to node :math:`j` in the
        second layer.  Vice versa, :math:`CA_{ji} = 1` indicates that node
        :math:`j` in the first layer is linked to node :math:`i` in the second
        layer.

        .. note::
           Cross adjacency matrix is **NEITHER** square **NOR** symmetric in
           general!

        :rtype: 2D Numpy array [index_1, index_2]
        :return: the cross adjacency matrix.
        """
        return self.cross_adjacency(node_list1=self.nodes_1,
                                    node_list2=self.nodes_2)

    def path_lengths_1(self, link_attribute=None):
        """
        Return internal path length matrix of first layer.

        Contains the paths length between all pairs of nodes within layer 1.
        However, the paths themselves will generally contain nodes from both
        layers. To avoid this and only consider paths lying within layer 1,
        do the following::

            net_1 = coupled_network.network_1()
            path_lengths_1 = net_1.path_lengths(link_attribute)

        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: 2D array [index_1, index_1]
        :return: the internal path length matrix of first layer.
        """
        return self.internal_path_lengths(node_list=self.nodes_1,
                                          link_attribute=link_attribute)

    def path_lengths_2(self, link_attribute=None):
        """
        Return internal path length matrix of second layer.

        Contains the path lengths between all pairs of nodes within layer 2.
        However, the paths themselves will generally contain nodes from both
        layers. To avoid this and only consider paths lying within layer 2,
        do the following::

            net_2 = coupled_network.network_2()
            path_lengths_2 = net_2.path_lengths(link_attribute)

        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: 2D array [index_2, index_2]
        :return: the internal path length matrix of second layer.
        """
        return self.internal_path_lengths(node_list=self.nodes_2,
                                          link_attribute=link_attribute)

    def cross_path_lengths(self, link_attribute=None):
        """
        Return cross path length matrix.

        Contains the path length between nodes from different layers. The paths
        contain nodes from both layers.

        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: 2D array [index_1, index_2]
        :return: the cross path length matrix.
        """
        return InteractingNetworks.\
            cross_path_lengths(self, node_list1=self.nodes_1,
                               node_list2=self.nodes_2,
                               link_attribute=link_attribute)

    def cross_link_distance(self):
        """
        Return cross link distance matrix.

        Contains the distance between nodes from different layers.

        :rtype: 2D array [index_1, index_2]
        :return: the cross link distance matrix.
        """
        return self.distance()[self.nodes_1, :][:, self.nodes_2]

    #
    #  Define scalar coupled network statistics
    #

    def number_cross_layer_links(self):
        """
        Return the number of links between the two layers.

        :return int: the number of links between nodes from different layers.
        """
        return self.number_cross_links(node_list1=self.nodes_1,
                                       node_list2=self.nodes_2)

    def number_internal_links(self):
        """
        Return the number of links within each layer.

        :rtype: (int, int)
        :return: the number of links within each layer.
        """
        n_links_1 = InteractingNetworks.number_internal_links(
            self, self.nodes_1)
        n_links_2 = InteractingNetworks.number_internal_links(
            self, self.nodes_2)

        return (n_links_1, n_links_2)

    def cross_link_density(self):
        """
        Return the density of links between the two layers.

        :return float: the density of links between the two layers.
        """
        return InteractingNetworks.cross_link_density(
            self, node_list1=self.nodes_1, node_list2=self.nodes_2)

    def internal_link_density(self):
        """
        Return the density of links within the two layers.

        :rtype: (float, float)
        :return: the density of links within the two layers.
        """
        density_1 = InteractingNetworks.\
            internal_link_density(self, self.nodes_1)
        density_2 = InteractingNetworks.\
            internal_link_density(self, self.nodes_2)

        return (density_1, density_2)

    def internal_global_clustering(self):
        """
        Return global clustering coefficients for each layer separately.

        Internal global clustering coefficients are calculated as mean values
        from the local clustering sequence of the whole coupled network. This
        implies that triangles spanning both layers will generally contribute
        to the internal clustering coefficients.

        To avoid this and consider only triangles lying within each layer::

            net_1 = coupled_network.network_1()
            clustering_1 = net_1.global_clustering()
            net_2 = coupled_network.network_2()
            clustering_2 = net_2.global_clustering()

        :rtype: (float, float)
        :return: the internal global clustering coefficients.
        """
        clustering_1 = InteractingNetworks.\
            internal_global_clustering(self, self.nodes_1)
        clustering_2 = InteractingNetworks.\
            internal_global_clustering(self, self.nodes_2)

        return (clustering_1, clustering_2)

    def cross_global_clustering(self):
        """
        Return global cross clustering for coupled network.

        The global cross clustering coefficient C_v gives the average
        probability, that two randomly drawn neighbors in layer 2 of node v in
        layer 1 are also neighbors and vice versa. It counts triangles having
        one vertex in layer 1 and two vertices in layer 2 and vice versa.

        :rtype: (float, float)
        :return: the cross global clustering coefficients.
        """
        cc_12 = InteractingNetworks.cross_global_clustering(
            self, node_list1=self.nodes_1, node_list2=self.nodes_2)
        cc_21 = InteractingNetworks.cross_global_clustering(
            self, node_list1=self.nodes_2, node_list2=self.nodes_1)

        return (cc_12, cc_21)

    def cross_transitivity(self):
        """
        Return cross transitivity for coupled network.

        The cross transitivity is the probability, that
        two randomly drawn neighbors in layer 2 of node v in layer 1 are also
        neighbors and vice versa. It counts triangles having one vertex in
        layer 1 and two vertices in layer 2 and vice versa. Cross transitivity
        tends to weight low cross degree vertices less strongly when compared
        to the global cross clustering coefficient (see [Newman2003]_).

        :rtype: (float, float)
        :return: the cross transitivities.
        """
        ct_12 = InteractingNetworks.cross_transitivity(
            self, node_list1=self.nodes_1, node_list2=self.nodes_2)
        ct_21 = InteractingNetworks.cross_transitivity(
            self, node_list1=self.nodes_2, node_list2=self.nodes_1)
        return (ct_12, ct_21)

    def cross_average_link_distance(self, reverse=False):
        """
        Return the cross average link distance

        The cross average link distance is the average link distance of each
        node of the first subnetwork to the nodes of the second subnetwork
        it is connected to. If reverse is set to True, the method calculates
        the average link distance of each node of the second subnetwork to the
        nodes of the first subnetwork.

        :arg bool reverse: Replace the subnetworks.

        :rtype: 1D Numpy array
        :return: the cross average link distances
        """
        if reverse:
            ax = 0
        else:
            ax = 1

        adj = self.cross_layer_adjacency()
        cld = self.cross_link_distance()
        return np.sum(adj*cld, axis=ax) / np.sum(adj, axis=ax)

    def cross_average_path_length(self, link_attribute=None):
        """
        Return cross average path length.

        Return the average (weighted) shortest path length between all pairs
        of nodes from different layers only.

        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :return float: the cross average path length.
        """
        return InteractingNetworks.cross_average_path_length(
            self, node_list1=self.nodes_1, node_list2=self.nodes_2,
            link_attribute=link_attribute)

    def internal_average_path_length(self, link_attribute=None):
        """
        Return internal average path length.

        Return the average (weighted) shortest path length between all pairs
        of nodes within each layer separately for which a path exists. Paths
        between nodes from different layers are not included in the averages!

        However, even if the end points lie within the same layer, the paths
        themselves will generally contain nodes from both layers. To avoid
        this and only consider paths lying within layer i, do the following::

            net_i = coupled_network.network_i()
            path_lengths_i = net_i.path_lengths(link_attribute)

        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: (float, float)
        :return: the internal average path length.
        """
        apl_1 = InteractingNetworks.internal_average_path_length(
            self, node_list=self.nodes_1, link_attribute=link_attribute)
        apl_2 = InteractingNetworks.internal_average_path_length(
            self, node_list=self.nodes_2, link_attribute=link_attribute)

        return (apl_1, apl_2)

    #
    #  Define local coupled network measures
    #

    def cross_degree(self):
        """
        Return the cross degree sequences.

        Gives the number of links from a specific node in one layer to the
        other layer.

        :rtype: tuple of two 1D arrays [index]
        :return: the cross degree sequences.
        """
        cross_degree_1 = InteractingNetworks.cross_degree(
            self, node_list1=self.nodes_1, node_list2=self.nodes_2)
        cross_degree_2 = InteractingNetworks.cross_degree(
            self, node_list1=self.nodes_2, node_list2=self.nodes_1)
        return (cross_degree_1, cross_degree_2)

    def internal_degree(self):
        """
        Return the internal degree sequences.

        Gives the number of links from a specific node to other nodes within
        the same layer.

        :rtype: tuple of two 1D arrays [index]
        :return: the internal degree sequences.
        """
        degree_1 = InteractingNetworks.internal_degree(
            self, node_list=self.nodes_1)
        degree_2 = InteractingNetworks.internal_degree(
            self, node_list=self.nodes_2)
        return (degree_1, degree_2)

    def cross_local_clustering(self):
        """
        Return local cross clustering for coupled network.

        The local cross clustering coefficient C_v gives the probability, that
        two randomly drawn neighbors in layer 2 of node v in layer 1 are also
        neighbors and vice versa. It counts triangles having one vertex in
        layer 1 and two vertices in layer 2 and vice versa.

        :rtype: tuple of two 1D arrays [index]
        :return: the cross local clustering coefficients.
        """
        cc_12 = InteractingNetworks.cross_local_clustering(
            self, node_list1=self.nodes_1, node_list2=self.nodes_2)
        cc_21 = InteractingNetworks.cross_local_clustering(
            self, node_list1=self.nodes_2, node_list2=self.nodes_1)
        return (cc_12, cc_21)

    def cross_closeness(self, link_attribute=None):
        """
        Return cross closeness sequence.

        Gives the inverse average geodesic distance from a node in one layer
        to all nodes in the other layer.

        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: tuple of two 1D arrays [index]
        :return: the cross closeness sequence.
        """
        cc_12 = InteractingNetworks.cross_closeness(
            self, node_list1=self.nodes_1, node_list2=self.nodes_2,
            link_attribute=link_attribute)
        cc_21 = InteractingNetworks.cross_closeness(
            self, node_list1=self.nodes_2, node_list2=self.nodes_1,
            link_attribute=link_attribute)
        return (cc_12, cc_21)

    def internal_closeness(self, link_attribute=None):
        """
        Return internal closeness sequence.

        Gives the inverse average geodesic distance from a node to all other
        nodes in the same layer.

        However, the included paths will generally contain nodes from both
        layers. To avoid this, do the following::

            net_i = coupled_network.network_i()
            closeness_i = net_i.closeness(link_attribute)

        :arg str link_attribute: Optional name of the link attribute to be used
            as the links' length. If None, links have length 1. (Default: None)
        :rtype: tuple of two 1D arrays [index]
        :return: the internal closeness sequence.
        """
        closeness_1 = InteractingNetworks.internal_closeness(
            self, node_list=self.nodes_1, link_attribute=link_attribute)
        closeness_2 = InteractingNetworks.internal_closeness(
            self, node_list=self.nodes_2, link_attribute=link_attribute)
        return (closeness_1, closeness_2)

    def cross_betweenness(self):
        """
        Return the cross betweenness sequence.

        Gives the normalized number of shortest paths only between nodes from
        **different** layers, in which a node :math:`i` is contained. This is
        equivalent to the inter-regional / inter-group betweenness with respect
        to layer 1 and layer 2.

        :rtype: tuple of two 1D arrays [index]
        :return: the cross betweenness sequence.
        """
        cb = InteractingNetworks.cross_betweenness(
            self, node_list1=self.nodes_1, node_list2=self.nodes_2)
        return (cb[self.nodes_1], cb[self.nodes_2])

    def internal_betweenness_1(self):
        """
        Return the internal betweenness sequences for layer 1.

        Gives the normalized number of shortest paths only between nodes from
        layer 1, in which a node :math:`i` is contained. :math:`i` can be
        member of any of the two layers. This is equivalent to the
        inter-regional / inter-group betweenness with respect to layer 1 and
        layer 1.

        :rtype: tuple of two 1D arrays [index]
        :return: the internal betweenness sequence for layer 1.
        """
        ib = self.internal_betweenness(self.nodes_1)

        return (ib[self.nodes_1], ib[self.nodes_2])

    def internal_betweenness_2(self):
        """
        Return the internal betweenness sequences for layer 2.

        Gives the normalized number of shortest paths only between nodes from
        layer 2, in which a node :math:`i` is contained. :math:`i` can be
        member of any of the two layers. This is equivalent to the
        inter-regional / inter-group betweenness with respect to layer 2 and
        layer 2.

        :rtype: tuple of two 1D arrays [index]
        :return: the internal betweenness sequence for layer 2.
        """
        ib = self.internal_betweenness(self.nodes_2)

        return (ib[self.nodes_1], ib[self.nodes_2])



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

cimport cython

import numpy as np
cimport numpy as cnp
from numpy cimport ndarray

from ...core._ext.types import FIELD, DFIELD
from ...core._ext.types cimport MASK_t, FIELD_t, DFIELD_t

cdef extern from "src_numerics.c":
    void _mutual_information(
            float *anomaly, int n_samples, int N, int n_bins, double scaling,
            double range_min, long *symbolic, long *hist, long *hist2d,
            float *mi)
    void _spearman_corr(int m, int tmax, bint *final_mask,
            float *time_series_ranked, float *spearman_rho)


# mutual_info =================================================================


def mutual_information(
    ndarray[FIELD_t, ndim=2, mode='c'] anomaly not None,
    int n_samples, int N, int n_bins, float scaling, float range_min):

    cdef:
        ndarray[long, ndim=2, mode='c'] symbolic = np.zeros(
            (N, n_samples), dtype=long)
        ndarray[long, ndim=2, mode='c'] hist = np.zeros(
            (N, n_bins), dtype=long)
        ndarray[long, ndim=2, mode='c'] hist2d = np.zeros(
            (n_bins, n_bins), dtype=long)
        ndarray[FIELD_t, ndim=2, mode='c'] mi = np.zeros(
            (N, N), dtype=FIELD)

    _mutual_information(
        <FIELD_t*> cnp.PyArray_DATA(anomaly), n_samples, N, n_bins, scaling,
        range_min, <long*> cnp.PyArray_DATA(symbolic),
        <long*> cnp.PyArray_DATA(hist), <long*> cnp.PyArray_DATA(hist2d),
        <FIELD_t*> cnp.PyArray_DATA(mi))

    return mi


# rainfall ====================================================================


def spearman_corr(int m, int tmax,
    ndarray[MASK_t, ndim=2, mode='c'] final_mask not None,
    ndarray[FIELD_t, ndim=2, mode='c'] time_series_ranked not None):

    cdef ndarray[FIELD_t, ndim=2, mode='c'] spearman_rho = np.zeros(
        (m, m), dtype=FIELD)

    _spearman_corr(m, tmax,
            <bint*> cnp.PyArray_DATA(final_mask),
            <FIELD_t*> cnp.PyArray_DATA(time_series_ranked),
            <FIELD_t*> cnp.PyArray_DATA(spearman_rho))

    return spearman_rho



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"
"""
Simple tests for the GeoGrid class.
"""
import numpy as np

from pyunicorn.core.geo_grid import GeoGrid


def test_RegularGrid():
    res = GeoGrid.RegularGrid(time_seq=np.arange(2),
                              space_grid=(np.array([0., 5.]),
                                          np.array([1., 2.])),
                              silence_level=2).lat_sequence()
    exp = np.array([0., 0., 5., 5.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)

    res = GeoGrid.RegularGrid(time_seq=np.arange(2),
                              space_grid=(np.array([0., 5.]),
                                          np.array([1., 2.])),
                              silence_level=2).lon_sequence()
    exp = np.array([1., 2., 1., 2.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_coord_sequence_from_rect_grid():
    res = GeoGrid.coord_sequence_from_rect_grid(lat_grid=np.array([0., 5.]),
                                                lon_grid=np.array([1., 2.]))
    exp = (np.array([0., 0., 5., 5.]), np.array([1., 2., 1., 2.]))
    assert np.allclose(res, exp, atol=1e-04)


def test_lat_sequence():
    res = GeoGrid.SmallTestGrid().lat_sequence()
    exp = np.array([0., 5., 10., 15., 20., 25.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_lon_sequence():
    res = GeoGrid.SmallTestGrid().lon_sequence()
    exp = np.array([2.5, 5., 7.5, 10., 12.5, 15.], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_convert_lon_coordinates():
    res = GeoGrid.SmallTestGrid().convert_lon_coordinates(
        np.array([10., 350., 20., 340., 170., 190.]))
    exp = np.array([10., -10., 20., -20., 170., -170.])
    assert np.allclose(res, exp, atol=1e-04)


def test_node_number2d():
    res = GeoGrid.SmallTestGrid().node_number(lat_node=14., lon_node=9.)
    exp = 3
    assert res == exp


def test_cos_lat():
    res = GeoGrid.SmallTestGrid().cos_lat()[:2]
    exp = np.array([1., 0.9962])
    assert np.allclose(res, exp, atol=1e-04)


def test_sin_lat():
    res = GeoGrid.SmallTestGrid().sin_lat()[:2]
    exp = np.array([0., 0.0872])
    assert np.allclose(res, exp, atol=1e-04)


def test_cos_lon():
    res = GeoGrid.SmallTestGrid().cos_lon()[:2]
    exp = np.array([0.999, 0.9962])
    assert np.allclose(res, exp, atol=1e-04)


def test_sin_lon():
    res = GeoGrid.SmallTestGrid().sin_lon()[:2]
    exp = np.array([0.0436, 0.0872])
    assert np.allclose(res, exp, atol=1e-04)


def test_angular_distance():
    res = GeoGrid.SmallTestGrid().angular_distance().round(2)
    exp = np.array([[0., 0.1, 0.19, 0.29, 0.39, 0.48],
                    [0.1, 0., 0.1, 0.19, 0.29, 0.39],
                    [0.19, 0.1, 0., 0.1, 0.19, 0.29],
                    [0.29, 0.19, 0.1, 0., 0.1, 0.19],
                    [0.39, 0.29, 0.19, 0.1, 0., 0.1],
                    [0.48, 0.39, 0.29, 0.19, 0.1, 0.]], dtype=np.float32)
    assert np.allclose(res, exp, atol=1e-04)


def test_boundaries():
    res = GeoGrid.SmallTestGrid().print_boundaries()
    exp = "         time     lat     lon\n   min    0.0    0.00    2.50\n" + \
          "   max    9.0   25.00   15.00"
    assert res == exp


def test_geometric_distance_distribution():
    res = GeoGrid.SmallTestGrid().geometric_distance_distribution(3)[0]
    exp = np.array([0.3333, 0.4667, 0.2])
    assert np.allclose(res, exp, atol=1e-04)

    res = GeoGrid.SmallTestGrid().geometric_distance_distribution(3)[1]
    exp = np.array([0., 0.1616, 0.3231, 0.4847])
    assert np.allclose(res, exp, atol=1e-04)


def test_region_indices():
    res = GeoGrid.SmallTestGrid().region_indices(
        np.array([0., 0., 0., 11., 11., 11., 11., 0.])).astype(int)
    exp = np.array([0, 1, 1, 0, 0, 0])
    assert np.allclose(res, exp, atol=1e-04)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

from platform import system
from setuptools import setup, Extension

from Cython.Build import cythonize
import numpy as np


# ==============================================================================


win = system() == 'Windows'
c_args = {
    'include_dirs': [np.get_include()],
    'extra_compile_args': ['/O2'] if win else ['-O3', '-std=c99', '-Wall'],
    'define_macros': [('_GNU_SOURCE', None),
                      ('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]}
cy_args = {
    'language_level': '3str', 'embedsignature': True,
    'boundscheck': True, 'wraparound': False,
    'initializedcheck': True, 'nonecheck': True,
    'warn.unused': True, 'warn.unused_arg': False, 'warn.unused_result': False}


# ==============================================================================


setup(
    ext_modules=cythonize(
        [Extension(
            f'pyunicorn.{pkg}._ext.numerics',
            sources=[f'src/pyunicorn/{pkg}/_ext/numerics.pyx'],
            **c_args)
         for pkg in ['climate', 'core', 'funcnet', 'timeseries']],
        compiler_directives=cy_args,
        nthreads=4))
